#  Copyright (c) 2023, Manfred Moitzi
#  License: MIT License
from typing import cast
import pytest
from collections import Counter
import ezdxf
from ezdxf import xref, colors, const
from ezdxf.math import Vec2
from ezdxf.document import Drawing
from ezdxf.layouts import Paperspace
from ezdxf.tools.standards import setup_dimstyle
from ezdxf.render.arrows import ARROWS
from ezdxf.entities import (
    Polyline,
    Polyface,
    factory,
    Insert,
    Dimension,
    Hatch,
    Image,
    MultiLeader,
    BlockRecord,
    Underlay,
)


def forward_handles(doc, count: int) -> None:
    for _ in range(count):
        doc.entitydb.next_handle()


def document_has_no_errors(doc: Drawing) -> bool:
    auditor = doc.audit()
    return not auditor.has_issues


class TestLoadResourcesWithoutNamingConflicts:
    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        doc.layers.add("FIRST")
        doc.linetypes.add(  # see also: complex_line_type_example.py
            "SQUARE",
            pattern="A,.25,-.1,[132,ltypeshp.shx,x=-.1,s=.1],-.1,1",
            description="Square -[]-----[]-----[]-----[]----[]----",
            length=1.45,
        )
        doc.linetypes.add(
            "GAS",
            pattern='A,.5,-.2,["GAS",STANDARD,S=.1,U=0.0,X=-0.1,Y=-.05],-.25',
            description="Gas ----GAS----GAS----GAS----GAS----GAS----GAS--",
            length=1,
        )
        arial = doc.styles.add("ARIAL", font="Arial.ttf")
        arial.set_extended_font_data(family="Arial", italic=False, bold=True)
        doc.layers.add("SECOND", linetype="SQUARE")
        dimstyle = setup_dimstyle(
            doc, "EZ_M_100_H25_CM", style="ARIAL", name="TestDimStyle"
        )
        dimstyle.dxf.dimltype = "GAS"
        dimstyle.dxf.dimblk = ARROWS.dot

        material = doc.materials.new("ExoticBlue")
        material.dxf.ambient_color_value = colors.encode_raw_color((0, 0, 255))
        layer = doc.layers.add("Layer_with_material")
        layer.dxf.material_handle = material.dxf.handle
        return doc

    def test_loading_a_simple_layer(self, sdoc):
        """This is the basic test to load a simple entity like a layer into a new
        document. It is checked whether all required structures have been created.
        """
        tdoc = ezdxf.new()
        loader = xref.Loader(sdoc, tdoc)
        loader.load_layers(["first"])
        loader.execute()
        assert document_has_no_errors(tdoc) is True

        layer = tdoc.layers.get("first")

        assert layer is not sdoc.layers.get("first"), "expected a copy"
        assert layer.dxf.name == "FIRST", "expected the original layer name"
        assert layer.doc is tdoc, "bound to wrong document"
        assert layer.dxf.handle in tdoc.entitydb, "entity not in database"
        assert layer.dxf.owner == tdoc.layers.head.dxf.handle, "invalid owner handle"
        assert layer.dxf.material_handle == tdoc.materials["global"].dxf.handle

    def test_loading_a_shape_linetype(self, sdoc):
        """Load a complex linetype with shapes which requires to load the dependent
        shape-file entry too.
        """
        tdoc = ezdxf.new()
        # handles shouldn't be synchronized to the source document!
        forward_handles(tdoc, 7)
        assert (
            sdoc.styles.find_shx("ltypeshp.shx").dxf.font == "ltypeshp.shx"
        ), "expected ltypeshp.shx entry to exist in the source document"

        loader = xref.Loader(sdoc, tdoc)
        loader.load_linetypes(["square"])
        loader.execute()
        assert document_has_no_errors(tdoc) is True

        ltype = tdoc.linetypes.get("square")
        assert ltype.dxf.name == "SQUARE"
        # do not repeat more tests from test_loading_a_simple_layer()

        style = tdoc.styles.find_shx("ltypeshp.shx")
        assert style.dxf.font == "ltypeshp.shx"
        pattern_style_handle = ltype.pattern_tags.get_style_handle()
        assert pattern_style_handle != "0"
        assert (
            pattern_style_handle == style.dxf.handle
        ), "expected handle of shape-file 'ltypeshp.shx' as pattern style handle"

    def test_loading_a_text_linetype(self, sdoc):
        """Load a complex linetype which contains text, the handle to the text style
        should point to the STANDARD text style in the target document.
        """
        tdoc = ezdxf.new()
        # handles shouldn't be synchronized to the source document!
        forward_handles(tdoc, 11)
        loader = xref.Loader(sdoc, tdoc)
        loader.load_linetypes(["gas"])
        loader.execute()
        assert document_has_no_errors(tdoc) is True

        ltype = tdoc.linetypes.get("gas")
        assert ltype.dxf.name == "GAS"
        # do not repeat more tests from test_loading_a_simple_layer()

        style = tdoc.styles.get("STANDARD")
        pattern_style_handle = ltype.pattern_tags.get_style_handle()
        assert pattern_style_handle != "0"
        assert (
            pattern_style_handle == style.dxf.handle
        ), "expected handle of text style STANDARD as pattern style handle"

    def test_loading_layer_with_complex_linetype(self, sdoc):
        """Loading a layer which references a complex linetype that also requires
        loading of an additional text style.
        """
        tdoc = ezdxf.new()
        loader = xref.Loader(sdoc, tdoc)
        loader.load_layers(["second"])
        loader.execute()
        assert document_has_no_errors(tdoc) is True

        layer = tdoc.layers.get("second")
        assert layer.dxf.name == "SECOND", "expected the original layer name"

        # Test if required resources are loaded:
        ltype = tdoc.linetypes.get(layer.dxf.linetype)
        assert ltype.dxf.name == "SQUARE", "expected linetype SQUARE in target doc"
        assert tdoc.styles.find_shx("ltypeshp.shx") is not None

    def test_loading_a_text_style_with_extended_font_data(self, sdoc):
        """The extended font data is stored into XDATA section of the STYLE table entry."""
        tdoc = ezdxf.new()
        loader = xref.Loader(sdoc, tdoc)
        loader.load_text_styles(["arial"])
        loader.execute()
        assert document_has_no_errors(tdoc) is True

        arial = tdoc.styles.get("arial")
        assert arial.dxf.name == "ARIAL", "expected text style ARIAL in target doc"

        family, italic, bold = arial.get_extended_font_data()
        assert family == "Arial"
        assert italic is False
        assert bold is True

    def test_loading_dimstyle(self, sdoc):
        tdoc = ezdxf.new()
        loader = xref.Loader(sdoc, tdoc)
        loader.load_dim_styles(["TestDimStyle"])
        loader.execute()
        assert document_has_no_errors(tdoc) is True

        dimstyle = tdoc.dimstyles.get("TestDimStyle")
        assert dimstyle.dxf.name == "TestDimStyle"

        assert dimstyle.dxf.dimtxsty == "ARIAL"
        arial = tdoc.styles.get("arial")
        assert arial.dxf.name == "ARIAL", "expected text style ARIAL in target doc"

        assert dimstyle.dxf.dimltype == "GAS"
        ltype = tdoc.linetypes.get("GAS")
        assert ltype.dxf.name == "GAS", "expected linetype GAS in target doc"

        assert dimstyle.dxf.dimblk == ARROWS.dot
        # Note: ACAD arrow head blocks are created automatically at export in
        # DimStyle.set_blk_handle() if they do not exist

    def test_loading_layer_with_custom_default_material(self, sdoc):
        tdoc = ezdxf.new()
        loader = xref.Loader(sdoc, tdoc)
        loader.load_layers(["Layer_with_material"])
        loader.execute()
        assert document_has_no_errors(tdoc) is True

        assert (
            "ExoticBlue" in tdoc.materials
        ), "expected copied entry in MATERIAL collection in target doc"
        layer = tdoc.layers.get("Layer_with_material")
        handle = layer.dxf.material_handle
        material = tdoc.entitydb.get(handle)
        assert material.dxf.name == "ExoticBlue"
        ambient_color = material.dxf.ambient_color_value
        assert colors.decode_raw_color(ambient_color)[1] == (0, 0, 255)


class TestLoadEntities:
    @pytest.fixture
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        doc.filename = "xref"
        doc.layers.add("Layer0")
        doc.linetypes.add("LType0", [0.0])  # CONTINUOUS
        doc.appids.add("TEST_ID")
        return doc

    def test_load_plain_entity(self, sdoc):
        msp = sdoc.modelspace()
        msp.add_point((0, 0), dxfattribs={"layer": "Layer0", "linetype": "LType0"})

        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True

        assert tdoc.layers.has_entry("Layer0") is True
        assert tdoc.linetypes.has_entry("LType0") is True

        target_msp = tdoc.modelspace()
        assert len(target_msp) == 1
        point = target_msp[0]
        assert point.doc is tdoc, "wrong document assigment"
        assert point.dxf.owner == target_msp.block_record_handle, "wrong owner handle"
        assert point.dxf.layer == "Layer0", "layer attribute not copied"

    def test_load_entity_layer_without_layer_table_entry(self, sdoc):
        msp = sdoc.modelspace()
        msp.add_point((0, 0), dxfattribs={"layer": "POINT"})
        tdoc = ezdxf.new()
        xref.load_modelspace(
            sdoc, tdoc, conflict_policy=xref.ConflictPolicy.XREF_PREFIX
        )
        assert document_has_no_errors(tdoc) is True
        target_msp = tdoc.modelspace()
        point = target_msp[0]
        assert (
            point.dxf.layer == "POINT"
        ), "layer names without a layer table entry can not be rename"

    def test_load_entity_with_xdata(self, sdoc):
        msp = sdoc.modelspace()
        point0 = msp.add_point((0, 0), dxfattribs={"layer": "Layer0"})
        point1 = msp.add_point((0, 0), dxfattribs={"layer": "Layer0"})
        point0.set_xdata(
            "TEST_ID",
            [
                (1000, "some text"),
                (1005, point1.dxf.handle),
                (1005, "FEFE"),  # FEFE to nothing and should be replaced by handle "0"
            ],
        )

        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True

        assert tdoc.appids.has_entry("TEST_ID")
        tp0, tp1 = tdoc.modelspace()
        assert tp0.xdata is not None
        xdata = tp0.get_xdata("TEST_ID")
        assert xdata[0] == (1000, "some text")
        assert xdata[1] == (
            1005,
            tp1.dxf.handle,
        ), "expected handle of point1 to be mapped"
        assert xdata[2] == (
            1005,
            "0",
        ), "expected un-mappable handle to be 0"

    def test_load_entity_with_reactors(self, sdoc):
        msp = sdoc.modelspace()
        point0 = msp.add_point((0, 0), dxfattribs={"layer": "Layer0"})
        point1 = msp.add_point((0, 0), dxfattribs={"layer": "Layer0"})
        point0.set_reactors([point1.dxf.handle, "FEFE"])

        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True

        tp0, tp1 = tdoc.modelspace()
        reactors = tp0.reactors.get()
        assert len(reactors) == 1
        assert reactors[0] == tp1.dxf.handle

    def test_load_entity_with_extension_dict(self, sdoc):
        msp = sdoc.modelspace()
        point0 = msp.add_point((0, 0), dxfattribs={"layer": "Layer0"})
        xdict = point0.new_extension_dict()
        xdict.add_dictionary_var("Test0", "Content0")

        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True

        copy = tdoc.modelspace()[0]
        assert copy.has_extension_dict is True
        xdict = copy.get_extension_dict()
        assert xdict.dictionary.doc is tdoc
        assert (xdict.dictionary in tdoc.objects) is True

        dict_var = xdict["Test0"]
        assert dict_var.dxf.value == "Content0"
        assert (dict_var in tdoc.objects) is True


class TestLoadTextEntities:
    @pytest.fixture
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        doc.styles.add("ARIAL", font="Arial.ttf")
        return doc

    def test_load_text_entity(self, sdoc):
        msp = sdoc.modelspace()
        msp.add_text("MyText", dxfattribs={"style": "ARIAL"})
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True

        assert tdoc.styles.has_entry("ARIAL")

    def test_load_mtext_entity(self, sdoc):
        msp = sdoc.modelspace()
        msp.add_mtext("MyText", dxfattribs={"style": "ARIAL"})
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True

        assert tdoc.styles.has_entry("ARIAL")

    def test_attdef_with_embedded_mtext_entity(self, sdoc):
        msp = sdoc.modelspace()
        attdef = msp.add_attdef("TEST", insert=(0, 0))
        mtext = msp.add_mtext("TEST", dxfattribs={"style": "ARIAL"})
        attdef.embed_mtext(mtext)
        # hack: embed_mtext() copies the style of the mtext into attdef
        # this may not always be true for loaded entities:
        attdef.dxf.style = "Standard"

        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert tdoc.styles.has_entry("ARIAL") is True


def test_load_mtext_with_columns():
    sdoc = ezdxf.new("R2000")
    sdoc.styles.add("ARIAL1", font="Arial.ttf")
    sdoc.styles.add("ARIAL2", font="Arial.ttf")
    mtext = sdoc.modelspace().add_mtext_static_columns(
        ["mtext0", "column1", "column2"],
        width=5,
        gutter_width=1,
        height=5,
        dxfattribs={"style": "ARIAL1"},
    )
    mtext.columns.linked_columns[1].dxf.style = "ARIAL2"

    tdoc = ezdxf.new("R2000")
    xref.load_modelspace(sdoc, tdoc)
    assert document_has_no_errors(tdoc) is True

    assert tdoc.styles.has_entry("ARIAL1")
    assert tdoc.styles.has_entry("ARIAL2")

    loaded_mtext = tdoc.modelspace()[0]
    assert loaded_mtext.has_columns is True
    columns = loaded_mtext.columns.linked_columns
    assert len(columns) == 2
    assert all(
        factory.is_bound(column, tdoc) for column in columns
    ), "all columns (MTEXT) should be bound to tdoc"


class TestLoadLinkedEntities:
    @pytest.fixture
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        doc.layers.add("Layer0")
        doc.linetypes.add("LType0", [0.0])  # CONTINUOUS
        return doc

    def test_load_polyline(self, sdoc):
        polyline = sdoc.modelspace().add_polyline2d([(0, 0), (1, 0), (1, 1)])
        polyline.vertices[0].dxf.linetype = "LType0"
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        # assert document_has_no_errors(tdoc) is True

        assert tdoc.linetypes.has_entry("LType0")
        loaded_polyline = cast(Polyline, tdoc.modelspace()[0])
        assert isinstance(loaded_polyline, Polyline)
        assert len(loaded_polyline.vertices) == 3
        assert all(
            factory.is_bound(v, tdoc) for v in loaded_polyline.vertices
        ), "all vertices should be bound to tdoc"
        assert all(
            v.dxf.owner == loaded_polyline.dxf.handle for v in loaded_polyline.vertices
        ), "all vertices are owned by the POLYLINE"
        assert (
            loaded_polyline.seqend.dxf.owner == loaded_polyline.dxf.handle
        ), "SEQEND is owned by the POLYLINE"

    def test_load_polyface(self, sdoc):
        polyface = sdoc.modelspace().add_polyface()
        polyface.append_face([(0, 0), (1, 0), (1, 1)])
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True

        loaded_polyface = cast(Polyface, tdoc.modelspace()[0])
        assert isinstance(loaded_polyface, Polyface)
        faces = list(loaded_polyface.faces())
        assert len(faces[0]) == 3 + 1  # vertices + face-record
        assert all(factory.is_bound(v, tdoc) for v in loaded_polyface.vertices)


class TestBlocks:
    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        doc.layers.add("Layer0")
        doc.linetypes.add("LType0", [0.0])  # CONTINUOUS
        block = doc.blocks.new("TestBlock")
        block.add_line(
            (0, 0), (1, 1), dxfattribs={"linetype": "LType0", "layer": "Layer0"}
        )
        block_ref = doc.modelspace().add_blockref("TestBlock", insert=(0, 0))
        block_ref.add_attrib("TestTag", "Content")
        return doc

    def test_load_block_layout(self, sdoc):
        tdoc = ezdxf.new()
        loader = xref.Loader(sdoc, tdoc)
        loader.load_block_layout(sdoc.blocks.get("TestBlock"))
        loader.execute()
        assert document_has_no_errors(tdoc) is True

        assert tdoc.layers.has_entry("Layer0"), "expected loaded LAYER"
        assert tdoc.linetypes.has_entry("LType0"), "expected loaded LTYPE"
        assert tdoc.block_records.has_entry("TestBlock"), "expected loaded BLOCK_RECORD"

        loaded_block = tdoc.blocks.get("TestBlock")
        assert loaded_block is not None, "loaded BlockLayout does not exist"
        assert len(loaded_block) == 1, "expected loaded block content"
        loaded_content = loaded_block[0]
        assert factory.is_bound(
            loaded_content, tdoc
        ), "invalid document binding of loaded block content"
        block_record = loaded_block.block_record
        assert (
            loaded_content.dxf.owner == block_record.dxf.handle
        ), "loaded content has invalid owner handle"
        assert (
            block_record.block.dxf.name == block_record.dxf.name
        ), "block name mismatch of BLOCK_RECORD and BLOCK"
        assert (
            factory.is_bound(block_record.block, tdoc) is True
        ), "BLOCK entity not bound to target doc"
        assert (
            factory.is_bound(block_record.endblk, tdoc) is True
        ), "ENDBLK entity not bound to target doc"

    def test_load_block_layout_does_type_checking(self, sdoc):
        tdoc = ezdxf.new()
        loader = xref.Loader(sdoc, tdoc)

        with pytest.raises(xref.LayoutError):
            loader.load_block_layout(sdoc.modelspace())

        with pytest.raises(xref.LayoutError):
            loader.load_block_layout(sdoc.paperspace())

        with pytest.raises(xref.LayoutError):
            loader.load_block_layout(None)

    def test_load_block_reference(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True

        assert (
            "TestBlock" in tdoc.blocks
        ), "expected TestBlock layout in target document"
        insert = cast(Insert, tdoc.modelspace()[0])
        assert isinstance(insert, Insert)
        assert factory.is_bound(insert, tdoc), "INSERT entity not bound to target doc"
        assert insert.dxf.name == "TestBlock"
        assert (
            insert.seqend.dxf.owner == insert.dxf.handle
        ), "SEQEND owner should be the INSERT entity"

    def test_load_block_reference_attributes(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True

        insert = cast(Insert, tdoc.modelspace()[0])
        assert len(insert.attribs) == 1, "expected loaded ATTRIB entities"

        attrib = insert.attribs[0]
        assert attrib.dxf.tag == "TestTag"
        assert attrib.dxf.text == "Content"

        assert factory.is_bound(attrib, tdoc), "ATTRIB entity not bound to target doc"
        assert (
            attrib.dxf.owner == insert.dxf.handle
        ), "ATTRIB owner should be the INSERT entity"


class TestAnonymousBlocks:
    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        anonymous_block = doc.blocks.new_anonymous_block("U")
        doc.modelspace().add_blockref(anonymous_block.dxf.name, insert=(0, 0))
        return doc

    def test_load_anonymous_block(self, sdoc):
        tdoc = ezdxf.new()
        for _ in range(3):  # increase anonymous block name counter
            tdoc.blocks.anonymous_block_name("U")
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True

        loaded_block_ref = cast(Insert, tdoc.modelspace()[0])
        loaded_block_name = loaded_block_ref.dxf.name
        assert (
            loaded_block_name == "*U4"
        ), "expected next available anonymous block name in tdoc"
        block_layout = tdoc.blocks.get(loaded_block_name)
        assert block_layout.block_record.dxf.name == loaded_block_name
        assert block_layout.block.dxf.name == loaded_block_name


def test_loaded_external_reference():
    sdoc = ezdxf.new()
    sdoc.add_xref_def("my.dxf", "my_xref")
    sdoc.modelspace().add_blockref("my_xref", insert=(0, 0))

    tdoc = ezdxf.new()
    xref.load_modelspace(sdoc, tdoc)

    assert tdoc.block_records.has_entry("my_xref")
    block_layout = tdoc.blocks.get("my_xref")
    assert block_layout.block.is_xref is True
    assert block_layout.block.dxf.xref_path == "my.dxf"

    insert = tdoc.modelspace()[0]
    assert isinstance(insert, Insert)
    assert insert.dxf.name == "my_xref"


def test_load_hard_owned_XRecord_within_appdata_section():
    """Any object which is hard-owned by another entity should automatically be loaded."""
    sdoc = ezdxf.new()
    line = sdoc.modelspace().add_line((0, 0), (1, 0))
    xrec = sdoc.objects.add_xrecord(line.dxf.handle)
    # The LINE entity owns the XRECORD by a hard-owner handle in the app-data
    # section "EZDXF":
    line.set_app_data("EZDXF", [(360, xrec.dxf.handle)])

    tdoc = ezdxf.new()
    xref.load_modelspace(sdoc, tdoc)
    assert document_has_no_errors(tdoc) is True

    loaded_line = tdoc.modelspace()[0]
    appdata = loaded_line.get_app_data("EZDXF")
    group_code, xrecord_handle = appdata[0]
    assert group_code == 360, "expected the group code of a hard-owner handle"
    loaded_xrecord = tdoc.entitydb.get(xrecord_handle)
    assert loaded_xrecord.dxf.owner == loaded_line.dxf.handle


def test_load_hard_owned_XRecord_by_extension_dict():
    """Any object of the extension dictionary is hard-owned by the entity and should
    automatically be loaded.
    """
    sdoc = ezdxf.new()
    line = sdoc.modelspace().add_line((0, 0), (1, 0))
    xdict = line.new_extension_dict()
    xrec = xdict.add_xrecord("EZDXF")
    xrec.extend([(3, "Content")])

    tdoc = ezdxf.new()
    xref.load_modelspace(sdoc, tdoc)
    assert document_has_no_errors(tdoc) is True

    loaded_line = tdoc.modelspace()[0]

    loaded_xdict = loaded_line.extension_dict
    assert loaded_xdict is not None
    assert (
        loaded_xdict.dictionary.dxf.owner == loaded_line.dxf.handle
    ), "Extension dictionary should be owned by the loaded entity"
    assert factory.is_bound(loaded_xdict.dictionary, tdoc)

    loaded_xrecord = loaded_xdict["EZDXF"]
    assert factory.is_bound(loaded_xrecord, tdoc)
    assert (
        loaded_xrecord.dxf.owner == loaded_xdict.dictionary.dxf.handle
    ), "XRecord should be owned by the extension dictionary"
    assert loaded_xrecord.tags[0] == (3, "Content")


class TestDimension:
    """Load a simple DIMENSION entity without DIMSTYLE overrides in the XDATA section."""

    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new(setup="dimstyles")
        msp = doc.modelspace()
        dim = msp.add_linear_dim(
            base=(3, 2),
            p1=(0, 0),
            p2=(3, 0),
            dimstyle="EZDXF",
        )
        dim.render()
        return doc

    def test_load_dimension_style_exist(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True
        assert tdoc.dimstyles.has_entry("EZDXF")

    def test_loaded_geometry_block_exist(self, sdoc):
        source_dim_block_name = sdoc.modelspace()[0].dxf.geometry
        source_block = sdoc.blocks.get(source_dim_block_name)
        assert len(source_block) == 9

        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True

        loaded_dim = tdoc.modelspace()[0]
        geometry = loaded_dim.dxf.geometry
        loaded_block = tdoc.blocks.get(geometry)
        assert len(loaded_block) == 9


class TestDimensionDimStyleOverride:
    """Load a DIMENSION with DIMSTYLE overrides in the XDATA section."""

    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new(setup="dimstyles")
        msp = doc.modelspace()
        dim = msp.add_linear_dim(
            base=(3, 2),
            p1=(0, 0),
            p2=(3, 0),
            dimstyle="EZDXF",
            override={"dimblk": ARROWS.dot},
        )
        dim.render()
        return doc

    def test_load_dimension_style_exist(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True
        assert tdoc.dimstyles.has_entry("EZDXF")

    def test_dot_blocks_in_source_doc(self, sdoc):
        source_dim_block_name = sdoc.modelspace()[0].dxf.geometry
        source_block = sdoc.blocks.get(source_dim_block_name)
        counter = Counter(
            block_ref.dxf.name for block_ref in source_block.query("INSERT")
        )
        assert counter["_DOT"] == 2

    def test_loaded_geometry_block_has_two_block_refs_of_dot(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True

        assert tdoc.block_records.has_entry("_DOT")
        loaded_dim = tdoc.modelspace()[0]
        geometry = loaded_dim.dxf.geometry
        loaded_block = tdoc.blocks.get(geometry)
        counter = Counter(
            block_ref.dxf.name for block_ref in loaded_block.query("INSERT")
        )
        assert counter["_DOT"] == 2

    def test_loaded_xdata_override_has_handle_to_existing_block(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True

        loaded_dim = cast(Dimension, tdoc.modelspace()[0])
        override_tags = loaded_dim.get_xdata_list("ACAD", "DSTYLE")

        handles = [value for code, value in override_tags if code == 1005]
        assert len(handles) == 1

        block_record = tdoc.entitydb.get(handles[0])
        assert block_record.dxf.name == "_DOT"


class TestLeader:
    """Load a LEADER with DIMSTYLE overrides in the XDATA section.

    Note: loading a LEADER entity does not automatically trigger the import of the
    linked TEXT, TOLERANCE or BLOCK entity, only handles are mapped correctly.

    """

    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new(setup="dimstyles")
        msp = doc.modelspace()
        text = msp.add_text("LEADER").set_placement((3, 1))
        msp.add_leader(
            vertices=[(0, 0), (1, 1), (2, 1)],
            dimstyle="EZDXF",
            override={"dimldrblk": ARROWS.dot},
            dxfattribs={"annotation_type": 0, "annotation_handle": text.dxf.handle},
        )
        return doc

    def test_loaded_leader_is_linked_to_loaded_text(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True

        text, leader = tdoc.modelspace()
        assert leader.dxf.annotation_handle == text.dxf.handle

    def test_loaded_xdata_override_has_handle_to_existing_block(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True

        text, leader = tdoc.modelspace()
        override_tags = leader.get_xdata_list("ACAD", "DSTYLE")

        handles = [value for code, value in override_tags if code == 1005]
        assert len(handles) == 1

        block_record = tdoc.entitydb.get(handles[0])
        assert block_record.dxf.name == "_DOT"


def test_tolerance_entity_register_dimstyle():
    from ezdxf.entities import Tolerance

    sdoc = ezdxf.new(setup="dimstyles")
    msp = sdoc.modelspace()

    # no factory method exist:
    tolerance = Tolerance.new(dxfattribs={"dimstyle": "EZDXF"}, doc=sdoc)
    msp.add_entity(tolerance)

    tdoc = ezdxf.new()
    xref.load_modelspace(sdoc, tdoc)
    assert document_has_no_errors(tdoc) is True
    assert tdoc.dimstyles.has_entry("EZDXF")


def test_associative_hatch_has_updated_source_boundary_handles():
    sdoc = ezdxf.new()
    msp = sdoc.modelspace()
    line0 = msp.add_line((0, 0), (2, 1))
    line1 = msp.add_line((2, 1), (1, 2))
    line2 = msp.add_line((1, 2), (0, 0))
    hatch = msp.add_hatch()
    path = hatch.paths.add_polyline_path([(0, 0), (2, 1), (1, 2)])
    path.source_boundary_objects = [e.dxf.handle for e in (line0, line1, line2)]

    tdoc = ezdxf.new()
    xref.load_modelspace(sdoc, tdoc)
    assert document_has_no_errors(tdoc) is True

    loaded_hatch = cast(Hatch, tdoc.modelspace().query("HATCH").first)
    loaded_handles = loaded_hatch.paths[0].source_boundary_objects

    assert len(loaded_handles) == 3
    assert all(h in tdoc.entitydb for h in loaded_handles)


class TestLoadImage:
    """Load a IMAGE, IMAGEDEF and IMAGEDEF_REACTOR"""

    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        msp = doc.modelspace()
        my_image_def = doc.add_image_def(
            filename="example.jpg", size_in_pixel=(640, 360)
        )
        msp.add_image(  # first image
            image_def=my_image_def, insert=(4, 5), size_in_units=(3.2, 1.8), rotation=30
        )
        msp.add_image(  # second image
            image_def=my_image_def, insert=(10, 5), size_in_units=(6.4, 3.6), rotation=0
        )
        return doc

    def test_loaded_infrastructure(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True
        assert "ACAD_IMAGE_VARS" in tdoc.rootdict, "expected required image vars"
        assert (
            len(tdoc.objects.query("IMAGEDEF_REACTOR")) == 2
        ), "expected two IMAGEDEF_REACTOR objects"
        assert len(tdoc.objects.query("IMAGEDEF")) == 1, "expected one IMAGEDEF object"

    def test_loaded_images_share_image_definition(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)

        image0, image1 = tdoc.modelspace()
        assert isinstance(image0, Image)
        assert isinstance(image1, Image)

        image0_def = image0.image_def
        assert image0_def.dxf.filename == "example.jpg"
        assert factory.is_bound(image0_def, tdoc)
        assert image0_def is image1.image_def

    def test_loaded_image_def_reactors(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)

        image0, image1 = tdoc.modelspace()
        assert isinstance(image0, Image)
        assert isinstance(image1, Image)

        # IMAGEDEF_REACTOR for image0
        assert factory.is_bound(image0.image_def_reactor, tdoc)
        assert image0.image_def_reactor.dxf.image_handle == image0.dxf.handle

        # IMAGEDEF_REACTOR for image1
        assert factory.is_bound(image1.image_def_reactor, tdoc)
        assert image1.image_def_reactor.dxf.image_handle == image1.dxf.handle

        # IMAGEDEF_REACTOR in IMAGEDEF reactor handles
        assert image0.image_def_reactor.dxf.handle in image0.image_def.get_reactors()
        assert image1.image_def_reactor.dxf.handle in image1.image_def.get_reactors()


class TestLoadWipeout:
    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        msp = doc.modelspace()
        msp.add_wipeout(vertices=[(0, 0), (2, 1), (1, 1)])
        return doc

    def test_loaded_infrastructure(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True
        assert "ACAD_WIPEOUT_VARS" in tdoc.rootdict, "expected required wipeout vars"

    def test_loaded_wipeout_has_same_boundary_path(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        source_wipeout = sdoc.modelspace()[0]
        loaded_wipeout = tdoc.modelspace()[0]

        assert all(
            v0.isclose(v1)
            for v0, v1 in zip(
                source_wipeout.boundary_path, loaded_wipeout.boundary_path
            )
        )
        # these handles are always "0"
        assert loaded_wipeout.dxf.image_def_reactor_handle == "0"
        assert loaded_wipeout.dxf.image_def_handle == "0"


class TestLoadMLine:
    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        style = doc.mline_styles.new("above")
        style.elements.append(0.5, 1)
        style.elements.append(0.25, 3)

        msp = doc.modelspace()
        msp.add_mline(
            [(0, 0), (10, 0), (15, 5), (15, 10)],
            dxfattribs={
                "style_name": style.dxf.name,
                "justification": const.MLINE_BOTTOM,
            },
        )
        return doc

    def test_loaded_infrastructure(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True
        assert "above" in tdoc.mline_styles

    def test_loaded_mline_has_correct_style_attributes(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)

        loaded_mline = tdoc.modelspace()[0]
        assert loaded_mline.style.dxf.name == "above"
        assert (
            loaded_mline.dxf.style_handle == tdoc.mline_styles.get("above").dxf.handle
        )

    def test_loaded_mline_has_same_vertices(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)

        source_mline = sdoc.modelspace()[0]
        loaded_mline = tdoc.modelspace()[0]
        assert all(
            v0.isclose(v1)
            for v0, v1 in zip(
                source_mline.get_locations(), loaded_mline.get_locations()
            )
        )


MTEXT_STYLE = "EZDXF_MTEXT"


class TestMultiLeaderMText:
    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        msp = doc.modelspace()

        doc.styles.add("OpenSans", font="OpenSans-Regular.ttf")
        mleader_style = doc.mleader_styles.duplicate_entry("Standard", MTEXT_STYLE)
        mleader_style.set_mtext_style("OpenSans")
        mleader_style.set_arrow_head(ARROWS.dot)
        mleader_style.set_leader_properties(linetype="CONTINUOUS")
        ml_builder = msp.add_multileader_mtext(MTEXT_STYLE)
        ml_builder.quick_leader(
            "MTEXT CONTENT", target=Vec2(40, 15), segment1=Vec2.from_deg_angle(45, 14)
        )
        return doc

    def test_loaded_infrastructure(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True
        assert tdoc.styles.has_entry("OpenSans") is True
        assert MTEXT_STYLE in tdoc.mleader_styles

    def test_loaded_mleader_mtext_style(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        loaded_mleader_style = tdoc.mleader_styles.get(MTEXT_STYLE)
        assert loaded_mleader_style.dxf.name == MTEXT_STYLE

        text_style = tdoc.styles.get("OpenSans")
        assert loaded_mleader_style.dxf.text_style_handle == text_style.dxf.handle

        ltype = tdoc.linetypes.get("CONTINUOUS")
        assert loaded_mleader_style.dxf.leader_linetype_handle == ltype.dxf.handle

        arrow_head_handle = loaded_mleader_style.dxf.arrow_head_handle
        arrow_head_block_record = tdoc.entitydb.get(arrow_head_handle)
        assert isinstance(arrow_head_block_record, BlockRecord)
        assert arrow_head_block_record.dxf.name == ARROWS.block_name(ARROWS.dot)

    def test_loader_multileader_attributes(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        db = tdoc.entitydb
        loaded_mleader: MultiLeader = tdoc.modelspace()[0]

        mleader_style = db.get(loaded_mleader.dxf.style_handle)
        assert mleader_style.dxf.name == MTEXT_STYLE, "invalid mleader_style_handle"

        ltype = db.get(loaded_mleader.dxf.leader_linetype_handle)
        assert ltype.dxf.name.upper() == "CONTINUOUS", "invalid linetype handle"

        text_style = db.get(loaded_mleader.dxf.text_style_handle)
        assert text_style.dxf.name == "OpenSans", "invalid text style handle"

        arrow_head = db.get(loaded_mleader.dxf.arrow_head_handle)
        assert arrow_head.dxf.name == ARROWS.block_name(
            ARROWS.dot
        ), "invalid arrowhead handle"

        assert (
            loaded_mleader.dxf.block_record_handle == "0"
        ), "expected null-ptr as block record handle"

        mtext_data = loaded_mleader.context.mtext
        assert (
            mtext_data.style_handle == text_style.dxf.handle
        ), "MTEXT data has invalid text style handle"

        assert mtext_data.default_content == "MTEXT CONTENT"


BLOCK_STYLE = "EZDXF_BLOCK"


class TestMultiLeaderBlock:
    @staticmethod
    def create_block(doc, name: str, size: float = 8, margin: float = 0.25):
        from ezdxf.render import forms
        from ezdxf.enums import TextEntityAlignment

        block = doc.blocks.new(name, base_point=(0, 0))
        block.add_lwpolyline(forms.square(size), close=True)
        attdef_attribs = {"height": 1.0, "style": "OpenSans"}
        bottom_left_attdef = block.add_attdef(
            "ONE", text="ONE", dxfattribs=attdef_attribs
        )
        bottom_left_attdef.set_placement(
            (margin, margin), align=TextEntityAlignment.BOTTOM_LEFT
        )
        top_right_attdef = block.add_attdef(
            "TWO", text="TWO", dxfattribs=attdef_attribs
        )
        top_right_attdef.set_placement(
            (size - margin, size - margin), align=TextEntityAlignment.TOP_RIGHT
        )
        return block

    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        from ezdxf.render import mleader

        doc = ezdxf.new()
        msp = doc.modelspace()

        doc.styles.add("OpenSans", font="OpenSans-Regular.ttf")
        block = self.create_block(doc, "SQUARE")

        mleader_style = doc.mleader_styles.duplicate_entry("Standard", BLOCK_STYLE)
        mleader_style.dxf.block_record_handle = block.block_record_handle

        ml_builder = msp.add_multileader_block(style=BLOCK_STYLE)
        ml_builder.set_content(name=block.name)
        ml_builder.set_attribute("ONE", "Data1")
        ml_builder.set_attribute("TWO", "Data2")
        ml_builder.add_leader_line(mleader.ConnectionSide.right, [Vec2(20, 10)])
        ml_builder.build(insert=Vec2(5, 2))
        return doc

    def test_loaded_mleader_block_style(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        loaded_mleader_style = tdoc.mleader_styles.get(BLOCK_STYLE)
        assert loaded_mleader_style.dxf.name == BLOCK_STYLE

        block_record_handle = loaded_mleader_style.dxf.block_record_handle
        loaded_block = tdoc.blocks.get("SQUARE")
        assert loaded_block.block_record_handle == block_record_handle

    def test_loaded_mleader_attributes(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        db = tdoc.entitydb
        loaded_mleader: MultiLeader = tdoc.modelspace()[0]
        block_record_handle = loaded_mleader.dxf.block_record_handle

        loaded_block_record = db.get(block_record_handle)
        assert isinstance(loaded_block_record, BlockRecord)
        assert loaded_block_record.dxf.name == "SQUARE"

        block_data = loaded_mleader.context.block
        assert block_data.block_record_handle == block_record_handle

    def test_loaded_block_attributes(self, sdoc):
        """Block attributes are virtual ATTRIB entities attached to the MLEADER BLOCK,
        created by the DXF renderer from ATTDEF entities in the BLOCK definition.

        """
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)

        loaded_mleader: MultiLeader = tdoc.modelspace()[0]
        block = tdoc.blocks.get("SQUARE")

        attdef0, attdef1 = block.query("ATTDEF")
        block_attrib0, block_attrib1 = loaded_mleader.block_attribs

        assert attdef0.dxf.handle == block_attrib0.handle
        assert attdef1.dxf.handle == block_attrib1.handle

        assert block_attrib0.text == "Data1"
        assert block_attrib1.text == "Data2"


class TestUnderlay:
    @pytest.fixture(scope="class")
    def sdoc(self) -> Drawing:
        doc = ezdxf.new()
        pdf_underlay_def = doc.add_underlay_def(
            filename="underlay.pdf", name="1"
        )  # name = page to display
        msp = doc.modelspace()
        msp.add_underlay(pdf_underlay_def, insert=(0, 0, 0), scale=1.0)
        msp.add_underlay(pdf_underlay_def, insert=(10, 0, 0), scale=2.0)
        return doc

    def test_loaded_infrastructure(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        assert document_has_no_errors(tdoc) is True
        assert "ACAD_PDFDEFINITIONS" in tdoc.rootdict
        acad_dict = tdoc.rootdict.get("ACAD_PDFDEFINITIONS")
        assert len(acad_dict) == 1, "expected one loaded pdf definition"

    def test_loaded_attributes(self, sdoc):
        tdoc = ezdxf.new()
        xref.load_modelspace(sdoc, tdoc)
        pdf_underlay0, pdf_underlay1 = tdoc.modelspace()
        assert isinstance(pdf_underlay0, Underlay)
        assert isinstance(pdf_underlay1, Underlay)

        pdf_definition = pdf_underlay0.get_underlay_def()
        assert pdf_definition.dxf.handle == pdf_underlay0.dxf.underlay_def_handle
        assert pdf_definition.dxf.handle == pdf_underlay1.dxf.underlay_def_handle
        assert pdf_underlay0.get_underlay_def() is pdf_underlay1.get_underlay_def()


class TestLoadPaperspaceLayout:
    @pytest.fixture(scope="class")
    def psp(self) -> Paperspace:
        doc = ezdxf.new()
        psp = doc.new_layout("MyLayout")
        # without a page setup no main viewport was created!
        psp.add_line((0, 0), (1, 0))
        return psp

    def test_loaded_infrastructure(self, psp):
        tdoc = ezdxf.new()
        xref.load_paperspace(psp, tdoc)
        assert document_has_no_errors(tdoc) is True

        loaded_psp = tdoc.paperspace("MyLayout")
        assert isinstance(loaded_psp, Paperspace)

    def test_loaded_paperspace_without_name_conflict(self, psp):
        tdoc = ezdxf.new()
        xref.load_paperspace(psp, tdoc)
        assert document_has_no_errors(tdoc) is True

        loaded_psp = tdoc.paperspace("MyLayout")
        assert isinstance(loaded_psp, Paperspace)

        assert loaded_psp.name == "MyLayout"
        assert loaded_psp.name in loaded_psp.doc.layouts

        block_record = loaded_psp.block_record
        assert block_record.dxf.name == "*Paper_Space0"
        assert loaded_psp.doc.block_records.has_entry("*Paper_Space0")

        # does the required block layout structure exist:
        block_layout = loaded_psp.doc.blocks.get("*Paper_Space0")
        assert block_layout.is_any_paperspace is True

        # check valid link structures
        dxf_layout = loaded_psp.dxf_layout
        assert block_record.dxf.layout == dxf_layout.dxf.handle
        assert dxf_layout.dxf.block_record_handle == block_record.dxf.handle

    def test_loaded_paperspace_content(self, psp):
        tdoc = ezdxf.new()
        xref.load_paperspace(psp, tdoc)
        loaded_psp = tdoc.paperspace("MyLayout")
        new_main_vp, line = loaded_psp
        assert line.dxftype() == "LINE"
        assert factory.is_bound(line, tdoc)
        # a new main VIEWPORT entity was created automatically:
        assert new_main_vp.dxftype() == "VIEWPORT"
        assert new_main_vp.dxf.id == 1

    def test_paperspace_name_conflict(self, psp):
        tdoc = ezdxf.new()
        tdoc.layouts.new("MyLayout")  # *Paper_Space0

        xref.load_paperspace(psp, tdoc)
        loaded_psp = tdoc.paperspace("MyLayout (2)")
        assert isinstance(loaded_psp, Paperspace)

        assert loaded_psp.name == "MyLayout (2)"
        assert loaded_psp.name in loaded_psp.doc.layouts

        block_record = loaded_psp.block_record
        assert block_record.dxf.name == "*Paper_Space1"
        assert loaded_psp.doc.block_records.has_entry("*Paper_Space1")


class TestLoadPaperspaceViewport:
    @pytest.fixture(scope="class")
    def psp(self) -> Paperspace:
        doc = ezdxf.new(setup=True)
        psp = doc.page_setup("MyLayout", fmt="ISO A3")
        ucs = doc.tables.ucs.add("MyUCS")
        visualstyle = doc.rootdict["ACAD_VISUALSTYLE"]["2dWireframe"]
        circle = psp.add_circle((100, 100), radius=50)
        vp = psp.add_viewport(
            center=(100, 100),
            size=(100, 100),
            view_center_point=(0, 0),
            view_height=10,
            status=2,
        )
        vp.dxf.clipping_boundary_handle = circle.dxf.handle
        vp.dxf.ucs_handle = ucs.dxf.handle
        vp.dxf.visual_style_handle = visualstyle.dxf.handle
        sun = factory.create_db_entry("SUN", {"owner": vp.dxf.handle}, doc)
        doc.objects.add_object(sun)
        vp.dxf.sun_handle = sun.dxf.handle
        return psp

    @pytest.fixture(scope="class")
    def loaded_psp(self, psp: Paperspace) -> Paperspace:
        tdoc = ezdxf.new()
        xref.load_paperspace(psp, tdoc)
        assert document_has_no_errors(tdoc) is True
        return tdoc.paperspace("MyLayout")

    def test_loaded_viewport_attributes(self, loaded_psp):
        tdoc = loaded_psp.doc
        loaded_vp = loaded_psp.viewports()[1]
        assert loaded_vp.dxf.center.isclose((100, 100))

        loaded_ucs = tdoc.tables.ucs.get("MyUCS")
        assert loaded_vp.dxf.ucs_handle == loaded_ucs.dxf.handle

        loaded_visual_style = tdoc.rootdict["ACAD_VISUALSTYLE"]["2dWireframe"]
        assert loaded_vp.dxf.visual_style_handle == loaded_visual_style.dxf.handle

    def test_loaded_clipping_entity(self, loaded_psp):
        loaded_vp = loaded_psp.viewports()[1]
        loaded_circle = loaded_psp.query("CIRCLE").first
        assert loaded_vp.dxf.clipping_boundary_handle == loaded_circle.dxf.handle

    def test_loaded_sun(self, loaded_psp):
        tdoc = loaded_psp.doc
        loaded_vp = loaded_psp.viewports()[1]
        loaded_sun = tdoc.entitydb.get(loaded_vp.dxf.sun_handle)
        assert factory.is_bound(loaded_sun, tdoc)


def test_load_empty_polyline_in_blk_ref():
    sdoc = ezdxf.new()
    blk = sdoc.blocks.new("TEST")
    blk.add_entity(Polyline())
    sdoc.modelspace().add_blockref("TEST", (0, 0))

    tdoc = ezdxf.new()
    xref.load_modelspace(sdoc, tdoc)

    blk = tdoc.blocks.get("TEST")
    assert len(blk) == 1
    p = blk[0]
    assert isinstance(p, Polyline) is True
    assert len(p) == 0


# TODO:
# DICTIONARY, test soft-ownership
# XRECORD, register and map pointers and hard-owner handles
# DXFTagStorage, register and map pointers and hard-owner handles
# SORTENTSTABLE
# GEODATA, map block_table_record, linked by extension dictionary of the modelspace
# ACIS entities
# ACAD_PROXY_ENTITY?
# OLE2FRAME?
# IDBUFFER/LAYERFILTER?

if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License
import pytest
from ezdxf.math import Vec2, Bezier4P
from ezdxf.render import hatching, forms
from ezdxf import path


class TestHatchBaseLine:
    def test_positive_line_distance(self):
        line = hatching.HatchBaseLine(
            origin=Vec2((1, 2)), direction=Vec2(2, 0), offset=Vec2(2, 2)
        )
        assert line.normal_distance == pytest.approx(2.0)

    def test_negative_line_distance(self):
        line = hatching.HatchBaseLine(
            origin=Vec2((1, 2)), direction=Vec2(2, 0), offset=Vec2(2, -2)
        )
        assert line.normal_distance == pytest.approx(-2.0)

    def test_hatch_line_direction_error(self):
        with pytest.raises(hatching.HatchLineDirectionError):
            hatching.HatchBaseLine(Vec2(), direction=Vec2(), offset=Vec2(1, 0))

    def test_dense_hatching_error(self):
        with pytest.raises(hatching.DenseHatchingLinesError):
            hatching.HatchBaseLine(
                Vec2(), direction=Vec2(1, 0), offset=Vec2(1, 0)
            )
        with pytest.raises(hatching.DenseHatchingLinesError):
            hatching.HatchBaseLine(
                Vec2(), direction=Vec2(1, 0), offset=Vec2(-1, 0)
            )

    def test_no_offset_error(self):
        with pytest.raises(hatching.DenseHatchingLinesError):
            hatching.HatchBaseLine(
                Vec2(), direction=Vec2(1, 0), offset=Vec2(0, 0)
            )

    def test_very_small_offset_error(self):
        with pytest.raises(hatching.DenseHatchingLinesError):
            hatching.HatchBaseLine(
                Vec2(), direction=Vec2(1, 0), offset=Vec2(0, 1e-6)
            )


class TestIntersectHatchLine:
    @pytest.fixture
    def horizontal_baseline(self):
        return hatching.HatchBaseLine(
            Vec2(), direction=Vec2(1, 0), offset=Vec2(0, 1)
        )

    def test_intersect_line_collinear(self, horizontal_baseline):
        a = Vec2(3, 0)
        b = Vec2(10, 0)
        distance = 0
        hatch_line = horizontal_baseline.hatch_line(distance)
        ip = hatch_line.intersect_line(a, b, distance, distance)
        assert ip.type == hatching.IntersectionType.COLLINEAR
        assert ip.p0 is a
        assert ip.p1 is b

    def test_intersect_line_start(self, horizontal_baseline):
        a = Vec2(0, 0)
        b = Vec2(0, 10)
        hatch_line = horizontal_baseline.hatch_line(0)
        ip = hatch_line.intersect_line(a, b, 0, 10)
        assert ip.type == hatching.IntersectionType.START
        assert ip.p0 is a

    def test_intersect_line_end(self, horizontal_baseline):
        a = Vec2(0, 0)
        b = Vec2(0, 10)

        hatch_line = horizontal_baseline.hatch_line(10)
        ip = hatch_line.intersect_line(a, b, 0, 10)
        assert ip.type == hatching.IntersectionType.END
        assert ip.p0 is b

    @pytest.mark.parametrize("d", [-2, 0, 6])
    def test_intersect_line_regular(self, horizontal_baseline, d):
        a = Vec2(4, -3)
        b = Vec2(4, 7)
        dist_a = horizontal_baseline.signed_distance(a)
        dist_b = horizontal_baseline.signed_distance(b)

        hatch_line = horizontal_baseline.hatch_line(d)
        ip = hatch_line.intersect_line(a, b, dist_a, dist_b)
        assert ip.type == hatching.IntersectionType.REGULAR
        assert ip.p0.isclose((4, d))

    def test_cubic_bezier_curve(self, horizontal_baseline):
        # low level intersection tests:
        # test_630b - TestRayCubicBezierCurve2dIntersection()
        curve = Bezier4P(Vec2.list([(0, -2), (2, 6), (4, -6), (6, 2)]))
        hatch_line = horizontal_baseline.hatch_line(0)
        ips = hatch_line.intersect_cubic_bezier_curve(curve)
        assert len(ips) == 3
        assert ips[0].p0.isclose((0.6762099922755492, 0.0))
        assert ips[1].p0.isclose((3.0, 0.0))
        assert ips[2].p0.isclose((5.323790007724451, 0.0))

    def test_missing_line_in_gear_example(self):
        baseline = hatching.HatchBaseLine(
            Vec2(), direction=Vec2(1, 1), offset=Vec2(-1, 1)
        )
        polygon = [
            Vec2(-5.099019513592784, 6.164414002968977),
            Vec2(-6.892024376045109, 7.245688373094721),
            Vec2(-7.245688373094716, 6.892024376045114),
            Vec2(-6.164414002968974, 5.099019513592788),
        ]
        lines = list(hatching.hatch_polygons(baseline, [polygon]))
        assert len(lines) == 2


DATA = [
    ("10 l 10 l 10", 10),
    ("2 l 2 r 2 r 2 l 6 " "l 10 l 2 l 2 r 2 r 2 l 6", 14),
    (
        "2 l 2 r 2 l 2 r 2 r 4 l 4 l 10 l 2 l 2 r 2 l 2 r 2 r 4 l 4",
        18,
    ),
    (
        "2 r 2 l 2 r 2 l 2 l 4 r 4 l 10 l 2 r 2 l 2 r 2 l 2 l 4 r 4",
        18,
    ),
    (
        "2 l 2 r 2 r 2 l 2 l 4 r 2 r 4 l 2 l 10 l 2 r 2 l 2 l 2 r 2 r 4 l 2 l 4 r 2",
        22,
    ),
    ("3 @2,2 @2,-2 3 l 10 l @-2,-2 @-2,2 2 @-2,-2 @-2,2", 14),
    (
        "3 @1,1 @1,1 @1,-1 @1,-1 3 l 10 l @-1,-1 @-1,-1 @-1,1 @-1,1 2 @-1,-1 @-1,-1 @-1,1 @-1,1",
        14,
    ),
]


@pytest.mark.parametrize("d,count", DATA)
def test_hatch_polygons(d: str, count):
    """Visual check by the function collinear_hatching() in script
    exploration/hatching.py,

    """
    baseline = hatching.HatchBaseLine(
        Vec2(), direction=Vec2(1, 0), offset=Vec2(0, 1)
    )
    lines = list(hatching.hatch_polygons(baseline, [list(forms.turtle(d))]))
    assert len(lines) == count


@pytest.mark.parametrize("d,count", DATA)
def test_hatch_paths(d: str, count):
    """Visual check by the function collinear_hatching() in script
    exploration/hatching.py,

    """
    baseline = hatching.HatchBaseLine(
        Vec2(), direction=Vec2(1, 0), offset=Vec2(0, 1)
    )
    lines = list(
        hatching.hatch_paths(
            baseline, [path.from_vertices(forms.turtle(d), close=True)]
        )
    )
    assert len(lines) == count


def test_hatch_curved_path():
    """Visual check by the function collinear_hatching() in script
    exploration/hatching.py,

    """
    p = path.Path((0, 0))
    p.line_to((10, 0))
    p.curve3_to((10, 10), (14, 5))
    p.line_to((0, 10))
    baseline = hatching.HatchBaseLine(
        Vec2(), direction=Vec2(1, 0), offset=Vec2(0, 1)
    )
    lines = list(hatching.hatch_paths(baseline, [p]))
    assert len(lines) == 10


def test_hatch_path_with_hole():
    baseline = hatching.HatchBaseLine(
        Vec2(), direction=Vec2(1, 0), offset=Vec2(0, 1)
    )
    polygons = [
        list(forms.square(10)),
        list(forms.translate(forms.square(6), (2, 2))),
    ]
    ctrL_lines = list(hatching.hatch_polygons(baseline, polygons))

    p = path.Path((0, 0))
    p.line_to((10, 0))
    p.curve3_to((10, 10), (14, 5))
    p.line_to((0, 10))
    p.close_sub_path()
    p.move_to((2, 2))
    p.line_to((8, 2))
    p.line_to((8, 8))
    p.line_to((2, 8))

    lines = list(hatching.hatch_paths(baseline, [p]))
    assert len(lines) == len(ctrL_lines)


def test_vertical_hatching_with_hole():
    """Visual check by the 1st example for 90 deg in function hole_examples()
    in script exploration/hatching.py,

    """
    size = 10
    angle = 90

    polygons = [
        list(forms.square(size)),
        list(forms.translate(forms.square(size - 2), (1, 1))),
        list(forms.translate(forms.square(3), (2, 2))),
        list(forms.translate(forms.square(3), (4, 3))),
    ]
    direction = Vec2.from_deg_angle(angle)
    offset = direction.orthogonal() * 0.1
    baseline = hatching.HatchBaseLine(
        Vec2(0, 0), direction=direction, offset=offset
    )
    lines = list(hatching.hatch_polygons(baseline, polygons))
    assert len(lines) == 241


class TestLinePatternRendering:
    @pytest.fixture
    def baseline(self):
        return hatching.HatchBaseLine(
            Vec2(),
            direction=Vec2(1, 0),
            offset=Vec2(0, 1),
            line_pattern=[1.0, -0.5, 1.5, -1.0],
        )

    def test_pattern_length(self, baseline):
        assert baseline.pattern_renderer(0).pattern_length == 4.0

    def test_render_full_pattern(self, baseline):
        renderer = baseline.pattern_renderer(0)
        lines = list(renderer.render_full_pattern(0))
        dash1, dash2 = lines
        assert dash1[0] == (0, 0)
        assert dash1[1] == (1, 0)
        assert dash2[0] == (1.5, 0)
        assert dash2[1] == (3.0, 0)

    def test_render_start_to_offset(self, baseline):
        renderer = baseline.pattern_renderer(0)
        lines = list(
            renderer.render_offset_to_offset(
                index=0, s_offset=0.0, e_offset=2.0
            )
        )
        dash1, dash2 = lines
        assert dash1[0] == (0, 0)
        assert dash1[1] == (1, 0)
        assert dash2[0] == (1.5, 0)
        assert dash2[1] == (2.0, 0)

    def test_render_offset_to_end(self, baseline):
        renderer = baseline.pattern_renderer(0)
        lines = list(
            renderer.render_offset_to_offset(
                index=0, s_offset=0.5, e_offset=4.0
            )
        )
        dash1, dash2 = lines
        assert dash1[0] == (0.5, 0)
        assert dash1[1] == (1, 0)
        assert dash2[0] == (1.5, 0)
        assert dash2[1] == (3.0, 0)

    def test_render_offset_to_offset(self, baseline):
        renderer = baseline.pattern_renderer(0)
        lines = list(
            renderer.render_offset_to_offset(
                index=0, s_offset=0.5, e_offset=2.0
            )
        )
        dash1, dash2 = lines
        assert dash1[0] == (0.5, 0)
        assert dash1[1] == (1, 0)
        assert dash2[0] == (1.5, 0)
        assert dash2[1] == (2.0, 0)

    def test_hatch_line_full_pattern(self, baseline):
        renderer = baseline.pattern_renderer(0)
        lines = list(renderer.render(Vec2(0, 0), Vec2(12, 0)))
        assert len(lines) == 6, "expected 3 full pattern sequences"
        assert lines[0][0] == (0, 0)
        assert lines[-1][1] == (11, 0)

    def test_hatch_line_with_start_and_end_offset(self, baseline):
        renderer = baseline.pattern_renderer(0)
        lines = list(renderer.render(Vec2(1, 0), Vec2(10, 0)))
        assert len(lines) == 6
        assert lines[0][0] == (1, 0)
        assert lines[-1][1] == (10, 0)


def test_explode_earth1_pattern():
    """Visual check by the function explode_hatch_pattern() in script
    exploration/hatching.py,

    """
    from ezdxf.entities import Hatch

    hatch = Hatch.new()
    hatch.set_pattern_definition(
        [
            [0.0, (0.0, 0.0), (1.5875, 1.5875), [1.5875, -1.5875]],
            [0.0, (0.0, 0.5953125), (1.5875, 1.5875), [1.5875, -1.5875]],
            [0.0, (0.0, 1.190625), (1.5875, 1.5875), [1.5875, -1.5875]],
            [
                90.0,
                (0.1984375, 1.3890625),
                (-1.5875, 1.5875),
                [1.5875, -1.5875],
            ],
            [90.0, (0.79375, 1.3890625), (-1.5875, 1.5875), [1.5875, -1.5875]],
            [
                90.0,
                (1.3890625, 1.3890625),
                (-1.5875, 1.5875),
                [1.5875, -1.5875],
            ],
        ]
    )
    hatch.dxf.solid_fill = 0
    # 1. polyline path
    hatch.paths.add_polyline_path(
        [
            (0.0, 223.0, 0.0),
            (10.0, 223.0, 0.0),
            (10.0, 233.0, 0.0),
            (0.0, 233.0, 0.0),
        ],
        is_closed=1,
        flags=3,
    )
    # jiggle_origin=True has random behavior, which is not good for a test!
    lines = list(hatching.hatch_entity(hatch, jiggle_origin=False))
    assert len(lines) == 139


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.math import is_axes_aligned_rectangle_2d, Vec2

@pytest.mark.parametrize(
    "points",
    [
        [(0, 0), (1, 0), (1, 1), (0, 1)],  # ccw start horizontal
        [(0, 0), (1, 0), (1, 1), (0, 1), (0, 0)],  # ccw start horizontal closed
        [(1, 0), (1, 1), (0, 1), (0, 0)],  # ccw start vertical
        [(1, 0), (1, 1), (0, 1), (0, 0), (1, 0)],  # ccw start vertical closed
        [(0, 1), (1, 1), (1, 0), (0, 0)],  # cw start horizontal
        [(0, 0), (0, 1), (1, 1), (1, 0)],  # cw start vertical
    ],
)
def test_is_aligned_rectangle(points):
    assert is_axes_aligned_rectangle_2d(Vec2.list(points)) is True


@pytest.mark.parametrize(
    "points",
    [
        [(0, 0), (1, 0), (1, 1)],  # less than 4 vertices
        [(0, 0), (1, 0), (1, 1), (0.1, 1)],  # nearly aligned quadliteral
        [
            (0, 0),
            (1, 0),
            (1.1, 1),
            (0.1, 1),
        ],  # parallelogram, opposed sides have the same length
    ],
)
def test_is_not_aligned_rectangle(points):
    assert is_axes_aligned_rectangle_2d(Vec2.list(points)) is False


if __name__ == '__main__':
    pytest.main([__file__])


# Copyright (c) 2018-2019 Manfred Moitzi
# License: MIT License
import pytest

from ezdxf.lldxf.types import DXFTag, get_xcode_for, is_valid_handle


def test_init():
    tag = DXFTag(1, "text")
    assert tag == (1, "text")

    tag2 = DXFTag(code=2, value="text2")
    assert tag2 == (2, "text2")


def test_immutability():
    tag = DXFTag(1, "text")
    with pytest.raises(AttributeError):
        tag.code = 2
    with pytest.raises(AttributeError):
        tag.value = "string"


def test_equality():
    assert (1, "text") == DXFTag(1, "text"), "should be equal to tuple"
    # Python 2.7 Issue
    assert ((1, "text") != DXFTag(1, "text")) is False

    assert (1, "text") != DXFTag(1, "text2"), "should be not equal to tuple"
    assert DXFTag(1, "text") == (1, "text"), "should be equal to tuple"
    assert DXFTag(1, "text") != (1, "text2"), "should be not equal to tuple"
    assert DXFTag(1, "text") == DXFTag(1, "text")
    assert DXFTag(1, "text") != DXFTag(1, "text2")


def test_index_able():
    tag = DXFTag(1, "text")
    assert tag[0] == 1
    assert tag[1] == "text"


def test_unpack():
    code, value = DXFTag(code=1, value="text")
    assert code == 1
    assert value == "text"


def test_iterable():
    tag = tuple(DXFTag(1, "text"))
    assert tag == (1, "text")


def test_public_attributes():
    tag = DXFTag(1, "text")
    assert tag.code == 1
    assert tag.value == "text"


def test_dxf_str():
    assert DXFTag(1, "text").dxfstr() == "  1\ntext\n"


def test_xcode_for():
    assert get_xcode_for(3) == 1000
    assert get_xcode_for(70) == 1070
    assert get_xcode_for(71) == 1070
    assert get_xcode_for(40) == 1040
    assert get_xcode_for(5) == 1005
    assert get_xcode_for(344) == 1005


def test_is_valid_handle():
    assert (
        is_valid_handle("0") is True
    ), "0 is a valid handle, but not allowed aa key in the entity database"
    assert is_valid_handle("ABBA") is True
    assert is_valid_handle("FEFE") is True
    assert is_valid_handle(None) is False, "None is not a valid handle"
    assert is_valid_handle("X") is False, "Not a valid hex value"
    assert is_valid_handle(1) is False, "Integer is not a valid handle"
    assert is_valid_handle(1.0) is False, "Float is not a valid handle"



# Copyright (c) 2019-2024 Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Optional, Iterator
from typing_extensions import Self
import logging
from ezdxf.lldxf import validator
from ezdxf.lldxf.attributes import (
    DXFAttr,
    DXFAttributes,
    DefSubclass,
    XType,
    RETURN_DEFAULT,
    group_code_mapping,
)
from ezdxf.lldxf.tags import Tags, DXFTag
from ezdxf.lldxf import const

from ezdxf.math import Vec3, UVec, X_AXIS, Z_AXIS, NULLVEC
from ezdxf.math.transformtools import transform_extrusion
from ezdxf.explode import explode_entity
from ezdxf.audit import AuditError
from .dxfentity import base_class, SubclassProcessor
from .dxfgfx import DXFGraphic, acdb_entity
from .factory import register_entity
from .dimension import OverrideMixin, register_override_handles
from .dimstyleoverride import DimStyleOverride
from .copy import default_copy

if TYPE_CHECKING:
    from ezdxf.audit import Auditor
    from ezdxf.entities import DXFNamespace, DXFEntity
    from ezdxf.layouts import BaseLayout
    from ezdxf.lldxf.tagwriter import AbstractTagWriter
    from ezdxf.math import Matrix44
    from ezdxf.query import EntityQuery
    from ezdxf import xref

__all__ = ["Leader"]
logger = logging.getLogger("ezdxf")

acdb_leader = DefSubclass(
    "AcDbLeader",
    {
        "dimstyle": DXFAttr(
            3,
            default="Standard",
            validator=validator.is_valid_table_name,
            # no fixer!
        ),
        # Arrowhead flag: 0/1 = no/yes
        "has_arrowhead": DXFAttr(
            71,
            default=1,
            optional=True,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        # Leader path type:
        # 0 = Straight line segments
        # 1 = Spline
        "path_type": DXFAttr(
            72,
            default=0,
            optional=True,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        # Annotation type or leader creation flag:
        # 0 = Created with text annotation
        # 1 = Created with tolerance annotation;
        # 2 = Created with block reference annotation
        # 3 = Created without any annotation
        "annotation_type": DXFAttr(
            73,
            default=3,
            validator=validator.is_in_integer_range(0, 4),
            fixer=RETURN_DEFAULT,
        ),
        # Hook line direction flag:
        # 1 = Hook line (or end of tangent for a spline leader) is the opposite
        #     direction from the horizontal vector
        # 0 = Hook line (or end of tangent for a spline leader) is the same
        #     direction as horizontal vector (see code 75)
        # DXF reference error: swapped meaning of 1/0
        "hookline_direction": DXFAttr(
            74,
            default=1,
            optional=True,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        # Hook line flag: 0/1 = no/yes
        "has_hookline": DXFAttr(
            75,
            default=1,
            optional=True,
            validator=validator.is_integer_bool,
            fixer=RETURN_DEFAULT,
        ),
        # Text annotation height:
        "text_height": DXFAttr(
            40,
            default=1,
            optional=True,
            validator=validator.is_greater_zero,
            fixer=RETURN_DEFAULT,
        ),
        # Text annotation width:
        "text_width": DXFAttr(
            41,
            default=1,
            optional=True,
            validator=validator.is_greater_zero,
            fixer=RETURN_DEFAULT,
        ),
        # 76: Number of vertices in leader (ignored for OPEN)
        # 10, 20, 30: Vertex coordinates (one entry for each vertex)
        # Color to use if leader's DIMCLRD = BYBLOCK
        "block_color": DXFAttr(
            77,
            default=7,
            optional=True,
            validator=validator.is_valid_aci_color,
            fixer=RETURN_DEFAULT,
        ),
        # Hard reference to associated annotation:
        # (mtext, tolerance, or insert entity)
        "annotation_handle": DXFAttr(340, default="0", optional=True),
        "normal_vector": DXFAttr(
            210,
            xtype=XType.point3d,
            default=Z_AXIS,
            optional=True,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
        # 'horizontal' direction for leader
        "horizontal_direction": DXFAttr(
            211,
            xtype=XType.point3d,
            default=X_AXIS,
            optional=True,
            validator=validator.is_not_null_vector,
            fixer=RETURN_DEFAULT,
        ),
        # Offset of last leader vertex from block reference insertion point
        "leader_offset_block_ref": DXFAttr(
            212, xtype=XType.point3d, default=NULLVEC, optional=True
        ),
        # Offset of last leader vertex from annotation placement point
        "leader_offset_annotation_placement": DXFAttr(
            213, xtype=XType.point3d, default=NULLVEC, optional=True
        ),
        # Xdata belonging to the application ID "ACAD" follows a leader entity if
        # any dimension overrides have been applied to this entity. See Dimension
        # Style Overrides.
    },
)
acdb_leader_group_codes = group_code_mapping(acdb_leader)


@register_entity
class Leader(DXFGraphic, OverrideMixin):
    """DXF LEADER entity"""

    DXFTYPE = "LEADER"
    DXFATTRIBS = DXFAttributes(base_class, acdb_entity, acdb_leader)
    MIN_DXF_VERSION_FOR_EXPORT = const.DXF2000

    def __init__(self) -> None:
        super().__init__()
        self.vertices: list[Vec3] = []

    def copy_data(self, entity: Self, copy_strategy=default_copy) -> None:
        """Copy vertices."""
        assert isinstance(entity, Leader)
        entity.vertices = Vec3.list(self.vertices)

    def load_dxf_attribs(
        self, processor: Optional[SubclassProcessor] = None
    ) -> DXFNamespace:
        dxf = super().load_dxf_attribs(processor)
        if processor:
            tags = processor.subclass_by_index(2)
            if tags:
                tags = Tags(self.load_vertices(tags))
                processor.fast_load_dxfattribs(
                    dxf, acdb_leader_group_codes, tags, recover=True
                )
            else:
                raise const.DXFStructureError(
                    f"missing 'AcDbLeader' subclass in LEADER(#{dxf.handle})"
                )

        return dxf

    def load_vertices(self, tags: Tags) -> Iterable[DXFTag]:
        for tag in tags:
            if tag.code == 10:
                self.vertices.append(tag.value)
            elif tag.code == 76:
                # Number of vertices in leader (ignored for OPEN)
                pass
            else:
                yield tag

    def preprocess_export(self, tagwriter: AbstractTagWriter) -> bool:
        if len(self.vertices) < 2:
            logger.debug(f"Invalid {str(self)}: more than 1 vertex required.")
            return False
        else:
            return True

    def export_entity(self, tagwriter: AbstractTagWriter) -> None:
        """Export entity specific data as DXF tags."""
        super().export_entity(tagwriter)
        tagwriter.write_tag2(const.SUBCLASS_MARKER, acdb_leader.name)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "dimstyle",
                "has_arrowhead",
                "path_type",
                "annotation_type",
                "hookline_direction",
                "has_hookline",
                "text_height",
                "text_width",
            ],
        )
        self.export_vertices(tagwriter)
        self.dxf.export_dxf_attribs(
            tagwriter,
            [
                "block_color",
                "annotation_handle",
                "normal_vector",
                "horizontal_direction",
                "leader_offset_block_ref",
                "leader_offset_annotation_placement",
            ],
        )

    def export_vertices(self, tagwriter: AbstractTagWriter) -> None:
        tagwriter.write_tag2(76, len(self.vertices))
        for vertex in self.vertices:
            tagwriter.write_vertex(10, vertex)

    def register_resources(self, registry: xref.Registry) -> None:
        """Register required resources to the resource registry."""
        assert self.doc is not None
        super().register_resources(registry)
        registry.add_dim_style(self.dxf.dimstyle)

        # The leader entity cannot register the annotation entity!
        if not self.has_xdata_list("ACAD", "DSTYLE"):
            return

        if self.doc.dxfversion > const.DXF12:
            # overridden resources are referenced by handle
            register_override_handles(self, registry)
        else:
            # overridden resources are referenced by name
            self.override().register_resources_r12(registry)

    def map_resources(self, clone: Self, mapping: xref.ResourceMapper) -> None:
        """Translate resources from self to the copied entity."""
        super().map_resources(clone, mapping)
        if self.dxf.hasattr("annotation_handle"):
            clone.dxf.annotation_handle = mapping.get_handle(self.dxf.annotation_handle)

        # DXF R2000+ references overridden resources by group code 1005 handles in the
        # XDATA section, which are automatically mapped by the parent class DXFEntity!
        assert self.doc is not None
        if self.doc.dxfversion > const.DXF12:
            return
        self_override = self.override()
        if not self_override.dimstyle_attribs:
            return  # has no overrides

        assert isinstance(clone, Leader)
        self_override.map_resources_r12(clone, mapping)

    def override(self) -> DimStyleOverride:
        """Returns the :class:`~ezdxf.entities.DimStyleOverride` object.

        .. warning::

            The LEADER entity shares only the DIMSTYLE override infrastructure with the
            DIMENSION entity but does not support any other features of the DIMENSION
            entity!

            HANDLE WITH CARE!

        """
        return DimStyleOverride(self)  # type: ignore

    def set_vertices(self, vertices: Iterable[UVec]):
        """Set vertices of the leader, vertices is an iterable of
        (x, y [,z]) tuples or :class:`~ezdxf.math.Vec3`.

        """
        self.vertices = [Vec3(v) for v in vertices]

    def transform(self, m: Matrix44) -> Leader:
        """Transform LEADER entity by transformation matrix `m` inplace."""
        self.vertices = list(m.transform_vertices(self.vertices))
        self.dxf.normal_vector, _ = transform_extrusion(
            self.dxf.normal_vector, m
        )  # ???
        self.dxf.horizontal_direction = m.transform_direction(
            self.dxf.horizontal_direction
        )
        self.post_transform(m)
        return self

    def __virtual_entities__(self) -> Iterator[DXFGraphic]:
        """Implements the SupportsVirtualEntities protocol."""
        from ezdxf.render.leader import virtual_entities

        for e in virtual_entities(self):
            e.set_source_of_copy(self)
            yield e

    def virtual_entities(self) -> Iterator[DXFGraphic]:
        """Yields the DXF primitives the LEADER entity is build up as virtual entities.

        These entities are located at the original location, but are not stored
        in the entity database, have no handle and are not assigned to any
        layout.
        """
        return self.__virtual_entities__()

    def explode(self, target_layout: Optional[BaseLayout] = None) -> EntityQuery:
        """Explode parts of the LEADER entity as DXF primitives into target layout,
        if target layout is ``None``, the target layout is the layout of the LEADER
        entity. This method destroys the source entity.

        Returns an :class:`~ezdxf.query.EntityQuery` container referencing all
        DXF primitives.

        Args:
            target_layout: target layout for the created DXF primitives, ``None`` for
                the same layout as the source entity.

        """
        return explode_entity(self, target_layout)

    def audit(self, auditor: Auditor) -> None:
        """Validity check."""
        super().audit(auditor)
        if len(self.vertices) < 2:
            auditor.fixed_error(
                code=AuditError.INVALID_VERTEX_COUNT,
                message=f"Deleted entity {str(self)} with invalid vertex count "
                f"= {len(self.vertices)}.",
                dxf_entity=self,
            )
            self.destroy()



#  Copyright (c) 2020, Manfred Moitzi
#  License: MIT License
import pytest
import os
import ezdxf
from ezdxf.audit import AuditError


def test_find_orphaned_layouts():
    doc = ezdxf.new()
    # Construct an orphaned LAYOUT entity:
    # Use proper owner, else we get audit error: INVALID_OWNER_HANDLE
    owner = doc.rootdict["ACAD_LAYOUT"].dxf.handle
    layout = doc.objects.new_entity(
        "LAYOUT", dxfattribs={"name": "Orphaned", "owner": owner}
    )

    auditor = doc.audit()
    assert len(auditor.fixes) == 1
    assert auditor.fixes[0].code == AuditError.ORPHANED_LAYOUT_ENTITY
    assert layout.is_alive is False


def test_find_orphaned_block_record():
    doc = ezdxf.new()
    # Construct an orphaned paperspace layout BLOCK_RECORD:
    doc.blocks.new("*Paper_Space99")
    auditor = doc.audit()
    assert len(auditor.fixes) == 1
    assert (
        auditor.fixes[0].code
        == AuditError.ORPHANED_PAPER_SPACE_BLOCK_RECORD_ENTITY
    )
    assert "*Paper_Space99" not in doc.blocks


@pytest.fixture
def MODEL_path() -> str:
    return os.path.join(os.path.dirname(__file__), "data", "MODEL.dxf")


def test_load_MODEL(MODEL_path):
    doc = ezdxf.readfile(MODEL_path)
    msp = doc.modelspace()
    assert msp.dxf.name == "MODEL", "Keep original uppercase name."
    assert "MODEL" in doc.rootdict["ACAD_LAYOUT"]
    assert "MODEL" in doc.layouts, "Check name as case insensitive string"
    assert "Model" in doc.layouts, "Check name as case insensitive string"


if __name__ == "__main__":
    pytest.main([__file__])



#  Copyright (c) 2022, Manfred Moitzi
#  License: MIT License

import pytest
from datetime import datetime
from ezdxf.acis import sab

T = sab.Tags


def test_decode_header(cube_sab):
    decoder = sab.Decoder(cube_sab)
    header = decoder.read_header()
    assert header.version == 21800
    assert header.n_records == 0
    assert header.n_entities == 2
    assert header.flags == 12
    assert header.product_id == "Open Design Alliance ACIS Builder"
    assert header.acis_version == "ACIS 218.00 NT"
    assert header.creation_date == datetime(2022, 5, 2, 5, 33, 25)
    assert header.units_in_mm == 1.0


def test_encode_header(cube_sab):
    decoder = sab.Decoder(cube_sab)
    header = decoder.read_header()
    data = header.dumpb()
    assert data == cube_sab[:len(data)]


def test_decode_first_record(cube_sab):
    decoder = sab.Decoder(cube_sab)
    _ = decoder.read_header()
    record = decoder.read_record()
    assert record == [
        (T.ENTITY_TYPE, "asmheader"),
        (T.POINTER, -1),
        (T.INT, -1),
        (T.STR, "208.0.4.7009"),
    ]


def test_decode_all_records(cube_sab):
    decoder = sab.Decoder(cube_sab)
    _ = decoder.read_header()
    records = list(decoder.read_records())
    assert len(records) == 116
    assert records[-1][0].value == "End-of-ASM-data"


def test_parse_sab(cube_sab):
    builder = sab.parse_sab(cube_sab)
    assert builder.header.version == 21800
    assert len(builder.entities) == 116
    assert builder.entities[0].name == "asmheader"
    assert builder.entities[-1].name == "End-of-ASM-data"


class TestSabEntity:
    @pytest.fixture(scope="class")
    def builder(self, cube_sab):
        return sab.parse_sab(cube_sab)

    @pytest.fixture(scope="class")
    def body(self, builder):
        return builder.bodies[0]

    def test_get_pointer_at_index(self, body):
        assert body.name == "body"
        assert body.attributes.is_null_ptr is False


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2022-2024, Manfred Moitzi
# License: MIT License
from __future__ import annotations
from typing import Iterable, Iterator, Sequence
import ezdxf
from ezdxf.math import Vec2, UVec, Vec3, safe_normal_vector, OCS

from ._mapbox_earcut import earcut

if ezdxf.options.use_c_ext:
    try:
        from ezdxf.acc.mapbox_earcut import earcut  # type: ignore
    except ImportError:
        pass

__all__ = [
    "mapbox_earcut_2d",
    "mapbox_earcut_3d",
]


def mapbox_earcut_2d(
    exterior: Iterable[UVec], holes: Iterable[Iterable[UVec]] | None = None
) -> list[Sequence[Vec2]]:
    """Mapbox triangulation algorithm with hole support for 2D polygons.

    Implements a modified ear slicing algorithm, optimized by z-order
    curve hashing and extended to handle holes, twisted polygons, degeneracies
    and self-intersections in a way that doesn't guarantee correctness of
    triangulation, but attempts to always produce acceptable results for
    practical data.

    Source: https://github.com/mapbox/earcut

    Args:
        exterior: exterior polygon as iterable of :class:`Vec2` objects
        holes: iterable of holes as iterable of :class:`Vec2` objects, a hole
            with single point represents a `Steiner point`_.

    Returns:
        yields the result as 3-tuples of :class:`Vec2` objects

    .. _Steiner point: https://en.wikipedia.org/wiki/Steiner_point_(computational_geometry)

    """
    points = Vec2.list(exterior)
    if len(points) == 0:
        return []
    holes_: list[list[Vec2]] = []
    if holes:
        holes_ = [Vec2.list(hole) for hole in holes]
    return earcut(points, holes_)


def mapbox_earcut_3d(
    exterior: Iterable[UVec], holes: Iterable[Iterable[UVec]] | None = None
) -> Iterator[Sequence[Vec3]]:
    """Mapbox triangulation algorithm with hole support for flat 3D polygons.

    Implements a modified ear slicing algorithm, optimized by z-order
    curve hashing and extended to handle holes, twisted polygons, degeneracies
    and self-intersections in a way that doesn't guarantee correctness of
    triangulation, but attempts to always produce acceptable results for
    practical data.

    Source: https://github.com/mapbox/earcut

    Args:
        exterior: exterior polygon as iterable of :class:`Vec3` objects
        holes: iterable of holes as iterable of :class:`Vec3` objects, a hole
            with single point represents a `Steiner point`_.

    Returns:
        yields the result as 3-tuples of :class:`Vec3` objects

    Raise:
        TypeError: invalid input data type
        ZeroDivisionError: normal vector calculation failed

    """
    polygon = Vec3.list(exterior)
    if len(polygon) == 0:
        return

    if polygon[0].isclose(polygon[-1]):
        polygon.pop()
    count = len(polygon)
    if count < 3:
        return
    if count == 3:
        yield polygon[0], polygon[1], polygon[2]
        return

    ocs = OCS(safe_normal_vector(polygon))
    elevation = ocs.from_wcs(polygon[0]).z
    exterior_ocs = list(ocs.points_from_wcs(polygon))
    holes_ocs: list[list[Vec3]] = []
    if holes:
        holes_ocs = [list(ocs.points_from_wcs(hole)) for hole in holes]

    # Vec3 supports the _Point protocol in _mapbox_earcut.py
    # required attributes: x, y
    for triangle in earcut(exterior_ocs, holes_ocs):
        yield tuple(
            ocs.points_to_wcs(Vec3(v.x, v.y, elevation) for v in triangle)
        )



# Copyright (c) 2015-2024, Manfred Moitzi
# License: MIT License
import pytest
import ezdxf


@pytest.fixture(scope="module")
def groups():
    doc = ezdxf.new("R2000")
    return doc.groups


def test_group_table_is_empty(groups):
    assert 0 == len(groups)


def test_create_new_group(groups):
    g = groups.new("MyGroup", description="The group description.")
    assert "GROUP" == g.dxftype()
    assert "MyGroup" in groups
    assert 0 == g.dxf.unnamed, "Named group has wrong unnamed attribute."
    assert 1 == g.dxf.selectable, "Group should be selectable by default."
    assert "The group description." == g.dxf.description
    assert 1 == len(groups)
    groups.clear()
    assert 0 == len(groups)


def test_create_unnamed_group(groups):
    g = groups.new()
    assert "GROUP" == g.dxftype()
    assert "*A1" in groups
    assert 1 == g.dxf.unnamed, "Unnamed group has wrong unnamed attribute."
    assert 1 == g.dxf.selectable, "Group should be selectable by default."
    assert "" == g.dxf.description, "Group description should be '' by default."
    assert 1 == len(groups)
    groups.clear()
    assert 0 == len(groups)


def test_delete_group_by_entity(groups):
    assert len(groups) == 0
    g = groups.new("MyGroup")
    assert len(groups) == 1
    groups.delete(g)
    assert len(groups) == 0


def test_delete_group_by_name(groups):
    assert "MyGroup" not in groups
    groups.new("MyGroup")
    groups.delete("MyGroup")
    assert 0 == len(groups)


def test_add_entities():
    dwg = ezdxf.new("R2000")
    group = dwg.groups.new()
    # the group itself is not an entity space, DXF entities has to be placed in model space, paper space
    # or in a block
    msp = dwg.modelspace()
    with group.edit_data() as e:  # e is a standard Python list of DXF entities
        line = msp.add_line((0, 0), (3, 0))
        e.append(line)
        e.append(msp.add_circle((0, 0), radius=2))

    assert len(group) == 2
    assert line in group

    group_handle = group.dxf.handle
    for entity in group:
        assert (
            group_handle in entity.reactors
        ), "expected group handle as reactor in entity"

    group.clear()
    assert len(group) == 0
    assert (
        group_handle not in line.reactors
    ), "group handle has to be removed from reactors"


def test_extend_group():
    dwg = ezdxf.new("R2000")
    group = dwg.groups.new()
    group_handle = group.dxf.handle
    # the group itself is not an entity space, DXF entities has to be placed in model space, paper space
    # or in a block
    msp = dwg.modelspace()
    line = msp.add_line((0, 0), (3, 0))
    group.set_data([line])
    assert len(group) == 1

    line2 = msp.add_line((0, 0), (3, 0))
    group.extend([line2])
    assert len(group) == 2
    assert group_handle in line2.reactors, "expected group handle as reactor in entity"

    group.extend([line])
    assert len(group) == 2, "cannot add same entity a second time"


if __name__ == "__main__":
    pytest.main([__file__])



# Copyright (c) 2024, Manfred Moitzi
# License: MIT License
from __future__ import annotations
import pytest

from ezdxf.math import Vec2, BoundingBox2d, is_point_in_polygon_2d
from ezdxf.math.clipping import InvertedClippingPolygon2d as ICP
from ezdxf.math.clipping import (
    find_closest_vertices,
    make_inverted_clipping_polygon,
)

# The InvertedClippingPolygon2d is based on the ConcaveClippingPolygon2d,
# so not all clipping tests are replicated here. See test_664.


@pytest.mark.parametrize(
    "extmin,extmax,expected",
    [
        [(0, 0), (1, 1), (0, 0)],
        [(8, 1), (9, 2), (1, 1)],
        [(8, 8), (9, 9), (2, 2)],
        [(1, 8), (2, 9), (3, 3)],
    ],
    ids=["lower-left", "lower-right", "upper-right", "upper-left"],
)
def test_find_closest_vertices(extmin, extmax, expected):
    outer = BoundingBox2d([(0, 0), (10, 10)]).rect_vertices()
    inner = BoundingBox2d([extmin, extmax]).rect_vertices()
    assert find_closest_vertices(inner, outer) == expected


# 9  7........6
# 8  ......3-2.
# 7  ......|.|.
# 6  ......0-1.
# 5  ..........
# 4  ..........
# 3  ..........
# 2  ..........
# 1  ..........
# 0  4........5
#    0123456789
def test_make_inverted_clipping_polygon():
    #                                4       6
    outer_boundary = BoundingBox2d([(0, 0), (9, 9)])
    #                           0       1       2       3
    inner_polygon = Vec2.list([(6, 6), (8, 6), (8, 8), (6, 8)])
    result = make_inverted_clipping_polygon(inner_polygon, outer_boundary)
    assert len(result) == 11
    # closest vertices are: 2<->6 where the the inner polygon will be connected
    # to the outer boundary
    for num, point in enumerate(
        [
            (8, 8),  # 2: start inner path at closest vertex pair
            (6, 8),  # 3: walk in counter clockwise order
            (6, 6),  # 0
            (8, 6),  # 1
            (8, 8),  # 2: close inner path
            (9, 9),  # 6: connect to outer boundary
            (9, 0),  # 5: walk in clockwise order
            (0, 0),  # 4
            (0, 9),  # 7
            (9, 9),  # 6: closer outer boundary
            (8, 8),  # 2: connect to inner polygon
        ]
    ):
        assert result[num] == point


# The connection between inner polygon and outer bounds, is created
# automatically at setup! This creates a closed concave clipping shape.
#
#    0123456789
# 9  ..........
# 8  .+------5.
# 7  .|......|.
# 6  .|.3--2.|.
# 5  .|.|..|.|.
# 4  .|.|..|.|.
# 3  .|.0--1.|.
# 2  .|/.....|.
# 1  .4------+.
# 0  ..........
#    0123456789
INNER_POLYGON = Vec2.list(
    # 0       1       2       3
    [(3, 3), (6, 3), (6, 6), (3, 6)]
)


@pytest.fixture(scope="module")
def inverted_polygon():
    #                                         4       5
    return ICP(INNER_POLYGON, BoundingBox2d([(1, 1), (8, 8)]))


#    0123456789
# 9  ..........
# 8  .+------5.
# 7  .|d....c|.
# 6  .|.3--2.|.
# 5  .|.|..|.|.
# 4  .|.|..|.|.
# 3  .|.0--1.|.
# 2  .|a....b|.
# 1  .4------+.
# 0  ..........
#    0123456789
#                           a       b       c       d
POINTS_INSIDE = Vec2.list([(2, 2), (7, 2), (7, 7), (2, 7)])

#    0123456789
# 9  g...f....e
# 8  .+------5.
# 7  .|......|.
# 6  .|.3--2.|.
# 5  .|.|..|.|.
# 4  h|.|i.|.|d
# 3  .|.0--1.|.
# 2  .|......|.
# 1  .4------+.
# 0  a...b....c
#    0123456789

POINTS_OUTSIDE = Vec2.list(
    # a       b       c       d       e       f       g       h       i
    [(0, 0), (4, 0), (9, 0), (9, 4), (9, 9), (4, 9), (0, 9), (0, 4), (4, 4)]
)


@pytest.mark.parametrize("point", POINTS_INSIDE)
def test_point_is_inside_polygon(point: Vec2, inverted_polygon: ICP):
    assert inverted_polygon.is_inside(point) is True


@pytest.mark.parametrize("point", POINTS_OUTSIDE)
def test_point_is_outside_polygon(point: Vec2, inverted_polygon: ICP):
    assert inverted_polygon.is_inside(point) is False


class TestLineClipping:
    # just test the basics to see if the implementation runs, more tests in test_664.
    def test_basic_clipping(self, inverted_polygon: ICP):
        """Does the inside/outside detection work for inverted polygons?"""
        # 5  .|.|..|.|.
        # 4  ax=x..x=xb
        # 3  .|.0--1.|.
        # 2  .|/.....|.
        # 1  .4------+.
        # 0  ..........
        #    0123456789
        result = inverted_polygon.clip_line(Vec2(0, 4), Vec2(9, 4))
        assert len(result) == 2
        l1, l2 = result
        assert l1[0].isclose((1, 4))
        assert l1[1].isclose((3, 4))
        assert l2[0].isclose((6, 4))
        assert l2[1].isclose((8, 4))

    def test_is_point_inside_inverted_polygon(self, inverted_polygon: ICP):
        polygon = inverted_polygon._clipping_polygon
        assert is_point_in_polygon_2d(Vec2(1.5, 2), polygon) == 1
        assert is_point_in_polygon_2d(Vec2(7.5, 2), polygon) == 1

    @pytest.mark.parametrize(
        "s,e,xs,xe",
        [
            [(0, 2), (9, 2), (1, 2), (8, 2)],  # outside, outside
            [(1, 2), (8, 2), (1, 2), (8, 2)],  # border, border
            [(1.5, 2), (7.5, 2), (1.5, 2), (7.5, 2)],  # inside, inside
        ],
        ids=["outside, outside", "border, border", "inside, inside"],
    )
    def test_connection_line_create_intersection_point(
        self, s, e, xs, xe, inverted_polygon: ICP
    ):
        """A line should be clipped at the diagonal connection line between 4<->0.
        There are two clipping lines 4->0 and 4<-0. The zero-length segment between the
        real segments should not be returned.

        inside, inside: #1101 <https://github.com/mozman/ezdxf/issues/1101>
        """
        # 5  .|.|..|.|.
        # 4  .|.|..|.|.
        # 3  .|.0--1.|.
        # 2  axx=====xb
        # 1  .4------+.
        # 0  ..........
        #    0123456789
        result = inverted_polygon.clip_line(Vec2(s), Vec2(e))
        assert len(result) == 2
        l1, l2 = result
        assert l1[0].isclose(xs)
        assert l1[1].isclose((2, 2))
        # zero-length segment (2, 2) -> (2, 2) is ignored or outside!
        assert l2[0].isclose((2, 2))
        assert l2[1].isclose(xe)

    def test_line_touches_vertex_at_coincident_edge(self, inverted_polygon: ICP):
        """Intersection point at coincident edge."""
        # 4  .|a|..|.|.
        # 3  .|.x--1.|.
        # 2  .|/.b...|.
        # 1  .4------+.
        # 0  ..........
        #    0123456789
        result = inverted_polygon.clip_line(Vec2(2, 4), Vec2(4, 2))
        assert len(result) == 2

        s, e = result[0]
        assert s.isclose((2, 4))
        assert e.isclose((3, 3))

        s, e = result[1]
        assert s.isclose((3, 3))
        assert e.isclose((4, 2))

    def test_line_touches_vertex_inside_inside(self, inverted_polygon: ICP):
        """No intersection point at vertex 1."""
        # 4  .|.|..|a|.
        # 3  .|.0--1.|.
        # 2  .|/..b..|.
        # 1  .4------+.
        # 0  ..........
        #    0123456789
        result = inverted_polygon.clip_line(Vec2(7, 4), Vec2(5, 2))
        assert len(result) == 1

        s, e = result[0]
        assert s.isclose((7, 4))
        assert e.isclose((5, 2))

    def test_colinear_line(self, inverted_polygon: ICP):
        """Colinear line segments are inside per definition."""
        # 5  .|.|..|.|.
        # 4  .|.|..|.|.
        # 3  ax=x==x=xb
        # 2  .|/.....|.
        # 1  .4------+.
        # 0  ..........
        #    0123456789
        result = inverted_polygon.clip_line(Vec2(0, 3), Vec2(9, 3))
        assert len(result) == 3
        l1, l2, l3 = result
        assert l1[0].isclose((1, 3))
        assert l1[1].isclose((3, 3))
        # zero-length segment (3, 3) -> (3, 3) is ignored or outside!
        assert l2[0].isclose((3, 3))
        assert l2[1].isclose((6, 3))
        assert l3[0].isclose((6, 3))
        assert l3[1].isclose((8, 3))


class TestPolygonClipping:
    # just test the basics to see if the implementation runs, more tests in test_664.
    def test_polygon_outside(self, inverted_polygon: ICP):
        # 9  ..........
        # 8  .+------5.
        # 7  .|......|.
        # 6  .|.3--2.|.
        # 5  .|.|dc|.|.
        # 4  .|.|ab|.|.
        # 3  .|.0--1.|.
        # 2  .|/.....|.
        # 1  .4------+.
        # 0  ..........
        #    0123456789

        result = inverted_polygon.clip_polygon(
            #           a       b       c       d
            Vec2.list([(4, 4), (5, 4), (5, 5), (4, 5)])
        )
        assert len(result) == 0

    def test_polygon_clipping(self, inverted_polygon: ICP):
        # 5  .|.|..|.|.
        # 4  .|.|..|.|.
        # 3  .|.0--1.|.
        # 2  .|/d--c.|.
        # 1  .4-x--x-+.
        # 0  ...a--b...
        #    0123456789

        result = inverted_polygon.clip_polygon(
            #           a       b       c       d
            Vec2.list([(3, 0), (6, 0), (6, 2), (3, 2)])
        )
        assert len(result) == 1
        clipped_polygon = result[0]
        # The clipped polygon doesn't have to be closed, future implementation may
        # change and the results of the Greiner-Horman algorithm for "edge-cases" are
        # not consistent!
        assert len(clipped_polygon) in (4, 5)
        if len(clipped_polygon) == 5:
            assert clipped_polygon[0].isclose(clipped_polygon[-1])
        bbox = BoundingBox2d(clipped_polygon)
        assert bbox.extmin.isclose((3, 1))  # lower-left clipped vertex
        assert bbox.extmax.isclose((6, 2))  # vertex c


if __name__ == "__main__":
    pytest.main([__file__])



# Created: 10.03.2011
# Copyright (c) 2011-2019, Manfred Moitzi
# License: MIT License
import pytest
from io import StringIO
from copy import deepcopy
from ezdxf.lldxf.tags import Tags, DXFTag
from ezdxf.lldxf.tagwriter import TagWriter
from ezdxf.lldxf.const import DXFValueError
from ezdxf.lldxf import types

TEST_TAGREADER = """  0
SECTION
  2
HEADER
  9
$ACADVER
  1
AC1018
  9
$DWGCODEPAGE
  3
ANSI_1252
  0
ENDSEC
  0
EOF
"""

TEST_TAGREADER_COMMENTS = """999
Comment0
  0
SECTION
  2
HEADER
  9
$ACADVER
999
Comment1
  1
AC1018
  9
$DWGCODEPAGE
  3
ANSI_1252
  0
ENDSEC
  0
EOF
"""

TESTHANDLE5 = """ 0
TEST
  5
F5
"""

TESTHANDLE105 = """ 0
TEST
105
F105
"""

TESTFINDALL = """  0
TEST0
  0
TEST1
  0
TEST2
"""

TAGS_WITH_VERTEX = """  0
TEST
 10
1.0
 20
2.0
 30
3.0
"""


class HandlesMock:
    calls = 0

    @property
    def next(self):
        self.calls += 1
        return "FF"


class TestTags:
    @pytest.fixture
    def tags(self):
        return Tags.from_text(TEST_TAGREADER)

    def test_from_text(self, tags):
        assert 8, len(tags)

    def test_write(self, tags):
        stream = StringIO()
        tagwriter = TagWriter(stream)
        tagwriter.write_tags(tags)
        result = stream.getvalue()
        stream.close()
        assert TEST_TAGREADER == result

    def test_update(self, tags):
        tags.update(DXFTag(2, "XHEADER"))
        assert "XHEADER" == tags[1].value

    def test_update_error(self, tags):
        with pytest.raises(DXFValueError):
            tags.update(DXFTag(999, "DOESNOTEXIST"))

    def test_set_first(self, tags):
        tags.set_first(DXFTag(999, "NEWTAG"))
        assert "NEWTAG" == tags[-1].value

    def test_find_first(self, tags):
        value = tags.get_first_value(9)
        assert "$ACADVER" == value

    def test_find_first_default(self, tags):
        value = tags.get_first_value(1234, default=999)
        assert 999 == value

    def test_find_first_error(self, tags):
        with pytest.raises(DXFValueError):
            tags.get_first_value(1234)

    def test_get_handle_5(self):
        tags = Tags.from_text(TESTHANDLE5)
        assert "F5" == tags.get_handle()

    def test_get_handle_105(self):
        tags = Tags.from_text(TESTHANDLE105)
        assert "F105" == tags.get_handle()

    def test_get_handle_create_new(self, tags):
        with pytest.raises(DXFValueError):
            tags.get_handle()

    def test_find_all(self):
        tags = Tags.from_text(TESTFINDALL)
        assert 3 == len(tags.find_all(0))

    def test_tag_index(self):
        tags = Tags.from_text(TESTFINDALL)
        index = tags.tag_index(0)
        assert 0 == index
        index = tags.tag_index(0, index + 1)
        assert 1 == index

    def test_find_first_value_error(self):
        tags = Tags.from_text(TESTFINDALL)
        with pytest.raises(DXFValueError):
            tags.tag_index(1)

    def test_clone_is_equal(self, tags):
        clone = tags.clone()
        assert id(tags) != id(clone)
        assert tags == clone

    def test_clone_is_independent(self, tags):
        clone = tags.clone()
        clone.pop()
        assert self.tags != clone

    def test_deepcopy(self):
        tags = Tags.from_text(TAGS_WITH_VERTEX)
        assert len(tags) == 2
        v = tags[1]
        assert v.value == (1.0, 2.0, 3.0)

        tags2 = deepcopy(tags)
        assert id(tags) != id(tags2)
        assert tags == tags2, "same content"
        # same ids, DXFTags are immutable
        assert id(v) == id(tags[1])

    def test_replace_handle_5(self):
        tags = Tags.from_text(TESTHANDLE5)
        tags.replace_handle("AA")
        assert "AA" == tags.get_handle()

    def test_replace_handle_105(self):
        tags = Tags.from_text(TESTHANDLE105)
        tags.replace_handle("AA")
        assert "AA" == tags.get_handle()

    def test_replace_no_handle_without_error(self, tags):
        tags.replace_handle("AA")
        with pytest.raises(DXFValueError):
            tags.get_handle()  # handle still doesn't exist

    def test_remove_tags(self, tags):
        tags.remove_tags(codes=(0,))
        assert 5 == len(tags)

    def test_strip_tags(self, tags):
        tags.remove_tags(codes=(0,))
        result = Tags.strip(tags, codes=(0,))
        assert 5 == len(result)
        assert isinstance(result, Tags)

    def test_has_tag(self, tags):
        assert tags.has_tag(2)

    def test_has_not_tag(self, tags):
        assert tags.has_tag(7) is False

    def test_pop_tags(self):
        tags = Tags(
            [
                DXFTag(1, "name1"),
                DXFTag(40, 1),
                DXFTag(40, 2),
                DXFTag(1, "name2"),
                DXFTag(40, 3),
                DXFTag(1, "name3"),
                DXFTag(40, 4),
                DXFTag(1, "name4"),
            ]
        )
        result = list(tags.pop_tags(codes=(40,)))
        assert len(result) == 4
        assert result[0] == (40, 1)
        assert result[-1] == (40, 4)

        assert len(tags) == 4
        assert tags[0] == (1, "name1")
        assert tags[-1] == (1, "name4")


class TestGetPointers:
    @pytest.fixture
    def tags(self) -> Tags:
        return Tags.from_tuples(
            [
                (0, "Entity"),
                # entity handle
                (5, "1"),
                # DIMSTYLE entity handle
                (105, "2"),
                # soft pointers - Translated during INSERT and XREF operations
                (330, "100"),
                (339, "101"),
                # soft owners - Translated during INSERT and XREF operations
                (350, "102"),
                (359, "103"),
                # hard pointers - Translated during INSERT and XREF operations
                (340, "200"),
                (349, "202"),
                # hard owners - Translated during INSERT and XREF operations
                (360, "203"),
                (369, "205"),
                # plotstylename (hard pointer) Translated during INSERT and XREF operations
                (390, "206"),
                (399, "207"),
                # hard pointers - Translated during INSERT and XREF operations
                (480, "208"),
                (481, "209"),
                # Arbitrary object handles; handle values that are taken "as is". They are not translated during INSERT and XREF operations
                (320, "300"),
                (329, "301"),
                # XDATA soft-pointer
                (1005, "400"),
            ]
        )

    def test_get_handle(self, tags):
        assert tags.get_handle() == "1"

    def test_get_soft_pointers(self, tags):
        assert tags.get_soft_pointers() == Tags.from_tuples(
            [
                (330, "100"),
                (339, "101"),
                (1005, "400"),
            ]
        )

    def test_get_soft_owners(self, tags):
        assert tags.get_soft_owner_handles() == Tags.from_tuples(
            [
                (350, "102"),
                (359, "103"),
            ]
        )

    def test_get_hard_pointers(self, tags):
        assert tags.get_hard_pointers() == Tags.from_tuples(
            [
                (340, "200"),
                (349, "202"),
                (390, "206"),
                (399, "207"),
                (480, "208"),
                (481, "209"),
            ]
        )

    def test_get_hard_owners(self, tags):
        assert tags.get_hard_owner_handles() == Tags.from_tuples(
            [
                (360, "203"),
                (369, "205"),
            ]
        )

    def test_get_translatable_pointers(self, tags):
        assert tags.get_hard_owner_handles() == Tags.from_tuples(
            [
                (360, "203"),
                (369, "205"),
            ]
        )

    def test_is_translatable_pointer(self, tags):
        assert tags.get_translatable_pointers() == Tags.from_tuples(
            [
                # soft pointers - Translated during INSERT and XREF operations
                (330, "100"),
                (339, "101"),
                # soft owners - Translated during INSERT and XREF operations
                (350, "102"),
                (359, "103"),
                # hard pointers - Translated during INSERT and XREF operations
                (340, "200"),
                (349, "202"),
                # hard owners - Translated during INSERT and XREF operations
                (360, "203"),
                (369, "205"),
                # plotstylename (hard pointer) Translated during INSERT and XREF operations
                (390, "206"),
                (399, "207"),
                # hard pointers - Translated during INSERT and XREF operations
                (480, "208"),
                (481, "209"),
                (1005, "400"),  # XDATA soft-pointer
            ]
        )

    def test_has_translatable_pointer(self):
        assert Tags.from_tuples([(5, "100")]).has_translatable_pointers() is False
        assert Tags.from_tuples([(330, "100")]).has_translatable_pointers() is True
        assert Tags.from_tuples([(320, "100")]).has_translatable_pointers() is False


DUPLICATETAGS = """  0
FIRST
  0
LAST
  1
TEST2
"""

COLLECT_1 = """  0
ZERO
  1
ONE
  2
TWO
  3
THREE
  4
FOUR
  0
ZERO
  1
ONE
  2
TWO
  3
THREE
  4
FOUR
"""


class TestTagsCollect:
    @pytest.fixture
    def tags(self):
        return Tags.from_text(COLLECT_1)

    def test_with_start_param(self, tags):
        collected_tags = tags.collect_consecutive_tags([1, 2, 3], start=1)
        assert 3 == len(collected_tags)
        assert "THREE" == collected_tags[2].value

    def test_with_end_param(self, tags):
        collected_tags = tags.collect_consecutive_tags([0, 1, 2, 3], end=3)
        assert 3 == len(collected_tags)
        assert "TWO" == collected_tags[2].value

    def test_with_start_and_end_param(self, tags):
        collected_tags = tags.collect_consecutive_tags([1, 2, 3], start=6, end=9)
        assert 3 == len(collected_tags)
        assert "THREE" == collected_tags[2].value

    def test_none_existing_codes(self, tags):
        collected_tags = tags.collect_consecutive_tags([7, 8, 9])
        assert 0 == len(collected_tags)

    def test_all_codes(self, tags):
        collected_tags = tags.collect_consecutive_tags([0, 1, 2, 3, 4])
        assert 10 == len(collected_tags)

    def test_emtpy_tags(self):
        tags = Tags()
        collected_tags = tags.collect_consecutive_tags([0, 1, 2, 3, 4])
        assert 0 == len(collected_tags)



# Copyright (c) 2020-2023, Matthew Broadway
# License: MIT License
from __future__ import annotations
from abc import ABC, abstractmethod, ABCMeta
from typing import Optional, Iterable

import numpy as np
from typing_extensions import TypeAlias
import dataclasses

from ezdxf.addons.drawing.config import Configuration
from ezdxf.addons.drawing.properties import Properties, BackendProperties
from ezdxf.addons.drawing.type_hints import Color
from ezdxf.entities import DXFGraphic
from ezdxf.math import Vec2, Matrix44
from ezdxf.npshapes import NumpyPath2d, NumpyPoints2d, single_paths

BkPath2d: TypeAlias = NumpyPath2d
BkPoints2d: TypeAlias = NumpyPoints2d

# fmt: off
_IMAGE_FLIP_MATRIX = [
    1.0, 0.0, 0.0, 0.0, 
    0.0, -1.0, 0.0, 0.0, 
    0.0, 0.0, 1.0, 0.0, 
    0.0, 999, 0.0, 1.0  # index 13: 999 = image height
]
# fmt: on


@dataclasses.dataclass
class ImageData:
    """Image data.

    Attributes:
        image: an array of RGBA pixels
        transform: the transformation to apply to the image when drawing
            (the transform from pixel coordinates to wcs)
        pixel_boundary_path: boundary path vertices in pixel coordinates, the image
            coordinate system has an inverted y-axis and the top-left corner is (0, 0)
        remove_outside: remove image outside the clipping boundary if ``True`` otherwise
            remove image inside the clipping boundary

    """

    image: np.ndarray
    transform: Matrix44
    pixel_boundary_path: NumpyPoints2d
    use_clipping_boundary: bool = False
    remove_outside: bool = True

    def image_size(self) -> tuple[int, int]:
        """Returns the image size as tuple (width, height)."""
        image_height, image_width, *_ = self.image.shape
        return image_width, image_height

    def flip_matrix(self) -> Matrix44:
        """Returns the transformation matrix to align the image coordinate system with
        the WCS.
        """
        _, image_height = self.image_size()
        _IMAGE_FLIP_MATRIX[13] = image_height
        return Matrix44(_IMAGE_FLIP_MATRIX)


class BackendInterface(ABC):
    """Public interface for 2D rendering backends."""

    @abstractmethod
    def configure(self, config: Configuration) -> None:
        raise NotImplementedError

    @abstractmethod
    def enter_entity(self, entity: DXFGraphic, properties: Properties) -> None:
        # gets the full DXF properties information
        raise NotImplementedError

    @abstractmethod
    def exit_entity(self, entity: DXFGraphic) -> None:
        raise NotImplementedError

    @abstractmethod
    def set_background(self, color: Color) -> None:
        raise NotImplementedError

    @abstractmethod
    def draw_point(self, pos: Vec2, properties: BackendProperties) -> None:
        raise NotImplementedError

    @abstractmethod
    def draw_line(self, start: Vec2, end: Vec2, properties: BackendProperties) -> None:
        raise NotImplementedError

    @abstractmethod
    def draw_solid_lines(
        self, lines: Iterable[tuple[Vec2, Vec2]], properties: BackendProperties
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def draw_path(self, path: BkPath2d, properties: BackendProperties) -> None:
        raise NotImplementedError

    @abstractmethod
    def draw_filled_paths(
        self, paths: Iterable[BkPath2d], properties: BackendProperties
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def draw_filled_polygon(
        self, points: BkPoints2d, properties: BackendProperties
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def draw_image(self, image_data: ImageData, properties: BackendProperties) -> None:
        raise NotImplementedError

    @abstractmethod
    def clear(self) -> None:
        raise NotImplementedError

    @abstractmethod
    def finalize(self) -> None:
        raise NotImplementedError


class Backend(BackendInterface, metaclass=ABCMeta):
    def __init__(self) -> None:
        self.entity_stack: list[tuple[DXFGraphic, Properties]] = []
        self.config: Configuration = Configuration()

    def configure(self, config: Configuration) -> None:
        self.config = config

    def enter_entity(self, entity: DXFGraphic, properties: Properties) -> None:
        # gets the full DXF properties information
        self.entity_stack.append((entity, properties))

    def exit_entity(self, entity: DXFGraphic) -> None:
        e, p = self.entity_stack.pop()
        assert e is entity, "entity stack mismatch"

    @property
    def current_entity(self) -> Optional[DXFGraphic]:
        """Obtain the current entity being drawn"""
        return self.entity_stack[-1][0] if self.entity_stack else None

    @abstractmethod
    def set_background(self, color: Color) -> None:
        raise NotImplementedError

    @abstractmethod
    def draw_point(self, pos: Vec2, properties: BackendProperties) -> None:
        """Draw a real dimensionless point, because not all backends support
        zero-length lines!
        """
        raise NotImplementedError

    @abstractmethod
    def draw_line(self, start: Vec2, end: Vec2, properties: BackendProperties) -> None:
        raise NotImplementedError

    def draw_solid_lines(
        self, lines: Iterable[tuple[Vec2, Vec2]], properties: BackendProperties
    ) -> None:
        """Fast method to draw a bunch of solid lines with the same properties."""
        # Must be overridden by the backend to gain a performance benefit.
        # This is the default implementation to ensure compatibility with
        # existing backends.
        for s, e in lines:
            if e.isclose(s):
                self.draw_point(s, properties)
            else:
                self.draw_line(s, e, properties)

    def draw_path(self, path: BkPath2d, properties: BackendProperties) -> None:
        """Draw an outline path (connected string of line segments and Bezier
        curves).

        The :meth:`draw_path` implementation is a fall-back implementation
        which approximates Bezier curves by flattening as line segments.
        Backends can override this method if better path drawing functionality
        is available for that backend.

        """
        if len(path):
            vertices = iter(
                path.flattening(distance=self.config.max_flattening_distance)
            )
            prev = next(vertices)
            for vertex in vertices:
                self.draw_line(prev, vertex, properties)
                prev = vertex

    def draw_filled_paths(
        self, paths: Iterable[BkPath2d], properties: BackendProperties
    ) -> None:
        """Draw multiple filled paths (connected string of line segments and
        Bezier curves).

        The current implementation passes these paths to the backend, all backends
        included in ezdxf handle holes by the even-odd method.  If a backend requires
        oriented paths (exterior paths in counter-clockwise and holes in clockwise
        orientation) use the function :func:`oriented_paths` to separate and orient the
        input paths.

        The default implementation draws all paths as filled polygons.

        Args:
            paths: sequence of paths
            properties: HATCH properties

        """
        for path in paths:
            self.draw_filled_polygon(
                BkPoints2d(
                    path.flattening(distance=self.config.max_flattening_distance)
                ),
                properties,
            )

    @abstractmethod
    def draw_filled_polygon(
        self, points: BkPoints2d, properties: BackendProperties
    ) -> None:
        """Fill a polygon whose outline is defined by the given points.
        Used to draw entities with simple outlines where :meth:`draw_path` may
        be an inefficient way to draw such a polygon.
        """
        raise NotImplementedError

    @abstractmethod
    def draw_image(self, image_data: ImageData, properties: BackendProperties) -> None:
        """Draw an image with the given pixels."""
        raise NotImplementedError

    @abstractmethod
    def clear(self) -> None:
        """Clear the canvas. Does not reset the internal state of the backend.
        Make sure that the previous drawing is finished before clearing.

        """
        raise NotImplementedError

    def finalize(self) -> None:
        pass


def oriented_paths(paths: Iterable[BkPath2d]) -> tuple[list[BkPath2d], list[BkPath2d]]:
    """Separate paths into exterior paths and holes. Exterior paths are oriented
    counter-clockwise, holes are oriented clockwise.
    """
    from ezdxf.path import winding_deconstruction, make_polygon_structure

    polygons = make_polygon_structure(single_paths(paths))
    external_paths: list[BkPath2d]
    holes: list[BkPath2d]
    external_paths, holes = winding_deconstruction(polygons)
    for p in external_paths:
        p.counter_clockwise()
    for p in holes:
        p.clockwise()
    return external_paths, holes



# Copyright (c) 2011-2020, Manfred Moitzi
# License: MIT License
import pytest

import ezdxf
from ezdxf.entities.textstyle import Textstyle, get_textstyle
from ezdxf.entities import DXFEntity


@pytest.fixture
def style():
    return Textstyle.new(
        "FFFF",
        dxfattribs={
            "name": "TEST",
            "font": "NOFONT.ttf",
            "width": 2.0,
        },
    )


def test_name(style):
    assert "TEST" == style.dxf.name


def test_font(style):
    assert "NOFONT.ttf" == style.dxf.font


def test_width(style):
    assert 2.0 == style.dxf.width


def test_height(style):
    assert 0.0 == style.dxf.height


def test_oblique(style):
    assert 0.0 == style.dxf.oblique


def test_bigfont(style):
    assert "" == style.dxf.bigfont


def test_is_backward(style):
    assert style.is_backward is False


def test_set_backward(style):
    style.is_backward = True
    assert style.is_backward is True


def test_is_upside_down(style):
    assert style.is_upside_down is False


def test_set_is_upside_down(style):
    style.is_upside_down = True
    assert style.is_upside_down is True


def test_set_is_vertical_stacked(style):
    style.is_vertical_stacked = True
    assert style.is_vertical_stacked is True


def test_not_existing_extended_font_data(style):
    assert style.has_extended_font_data is False
    assert style.get_extended_font_data() == ("", False, False)


@pytest.fixture
def xstyle():
    style = Textstyle.new(
        "FFFF",
        dxfattribs={
            "name": "OpenSans-BoldItalic",
            "font": "OpenSans-BoldItalic.ttf",
        },
    )
    style.set_xdata("ACAD", [(1000, "Open Sans"), [1071, 50331682]])
    return style


def test_extended_font_data(xstyle):
    assert xstyle.has_extended_font_data is True
    assert xstyle.get_extended_font_data() == ("Open Sans", True, True)


def test_discard_extended_font_data(xstyle):
    xstyle.discard_extended_font_data()
    assert xstyle.has_extended_font_data is False


def test_set_extended_font_data(style):
    style.set_extended_font_data("Arial", italic=True, bold=True)
    assert style.get_extended_font_data() == ("Arial", True, True)


def test_dxf_details_for_extended_font_data(style):
    style.set_extended_font_data("Arial", italic=True, bold=True)
    xdata = style.get_xdata("ACAD")
    assert len(xdata) == 2

    group_code, family = xdata[0]
    assert group_code == 1000
    assert family == "Arial"

    group_code, flags = xdata[1]
    assert group_code == 1071
    assert flags == 50331682


MALFORMED_STYLE = """0
STYLE
5
FEFE
100
AcDbTextStyleTableRecord
2
STY_EZDXF
70
0
40
0.0
41
1.5
50
30.0
71
0
42
0.0
3
txt
4

100
AcDbSymbolTableRecord
"""


def test_malformed_layer():
    style = Textstyle.from_text(MALFORMED_STYLE)
    assert style.dxf.name == "STY_EZDXF"
    assert style.dxf.handle == "FEFE"
    assert style.dxf.width == 1.5
    assert style.dxf.oblique == 30.0


@pytest.fixture(scope="module")
def msp():
    doc = ezdxf.new()
    style = doc.styles.add("MyStyle", font="arial.ttf")
    style.dxf.width = 2
    style.dxf.oblique = 75
    return doc.modelspace()


class TestGetTextStyle:
    def test_get_text_style_from_non_text_entity(self):
        style = get_textstyle(DXFEntity())
        assert style.dxf.name == "Standard"

    def test_get_undefined_text_style_from_text_entity(self, msp):
        text = msp.add_text("TEST", dxfattribs={"style": "Unknown"})
        style = get_textstyle(text)
        assert style.dxf.name == "Standard"

    def test_get_defined_text_style_from_text_entity(self, msp):
        text = msp.add_text("TEST", dxfattribs={"style": "MyStyle"})
        style = get_textstyle(text)
        assert style.dxf.name == "MyStyle"
        assert style.dxf.width == 2
        assert style.dxf.oblique == 75
